--[[

BattleGroundConstants.lua
	
	Contains constant values that are shared between the various battleground components.  Make
	sure all values are fully documented!

]]

local BattleGroundConstants = {}

-- General gameplay
BattleGroundConstants.BONUS_TURN_COUNT				= 4				-- number of gems matched in a single match to give a bonus turn
BattleGroundConstants.HEROIC_EFFORT_XP				= 100			-- amount of XP given to player on heroic effort
BattleGroundConstants.GEM_SWAP_TIME					= 400 --500			-- time taken to swap two gems when a move is made
BattleGroundConstants.SPIN_RATE						= math.pi		-- rate at which graphics will spin
BattleGroundConstants.FAST_SPIN_RATE				= 3 * math.pi	-- rate at which graphics will spin quickly

-- AI
BattleGroundConstants.AI_MOVE_DELAY					= 500			-- wait for a period from the start of a move before making decisions on AI (should be reduced on soak and regression tests)

-- Gem cascading configuration
BattleGroundConstants.DROP_DELAY					= 10 --100			-- the delay before a gem starts to drop (this value is multiplied by the gems current Y value)
BattleGroundConstants.DROP_XDELAY					= 5 --40			-- an extra delay before the gem starts to drop (multiplied by the gems current X value)
BattleGroundConstants.FALL_ACCELERATION				= 2000 --700		-- the gem falling acceleration, as passed to the movement controller
BattleGroundConstants.PAUSE_BETWEEN_CASCADES		= 100 --500			-- how long to pause between back to back cascades
BattleGroundConstants.PAUSE_AFTER_SPELL				= 200 --500			-- how long to pause after a spell finishes before checking for matches and cascading
BattleGroundConstants.PAUSE_AFTER_ITEM				= 200 --500			-- how long to pause after an item activation finishes before checking for matches and cascading, or ending turn
BattleGroundConstants.PAUSE_AFTER_MOVE				= 400 --500			-- how long the host should pause after making a swap to allow the player to see the move before cascades start
BattleGroundConstants.PAUSE_AFTER_MISS_MOVE			= 1000				-- how long to wait after the player forfeits their move and the turn indicate flips back (if the player was stunned, for example).
BattleGroundConstants.PAUSE_FOR_ITEM_EFFECT			= 2000				-- how long to wait for the item special effect to play before actually using that item


-- sorting values
BattleGroundConstants.BASE_SORTING_VALUE 			= 0
BattleGroundConstants.GEM_SELECTION_SORTING_VALUE 	= BattleGroundConstants.BASE_SORTING_VALUE - 2
BattleGroundConstants.HINT_ARROW_SORTING_VALUE 		= BattleGroundConstants.BASE_SORTING_VALUE - 3
BattleGroundConstants.FALLING_NUMBERS_SORTING_VALUE = BattleGroundConstants.BASE_SORTING_VALUE - 9
BattleGroundConstants.FX_CONTAINER_SORTING_VALUE 	= BattleGroundConstants.BASE_SORTING_VALUE - 10
BattleGroundConstants.HINT_ARROW_ALPHA_VALUE 		= 0.8

BattleGroundConstants.HEROIC_EFFORT_COUNT 			= 5				-- the number of cascades needed to get a heroic effort

BattleGroundConstants.WILDCARD_DIST					= {				-- the distribution of each kind of wildcard
	["2"] = 7,
	["3"] = 6,
	["4"] = 5,
	["5"] = 4,
	["6"] = 3,
	["7"] = 2,
}

BattleGroundConstants.MANA_POOL_NAMES 				= {
	"red", 
	"green", 
	"blue", 
	"yellow",
	"black",
	"power",
}

BattleGroundConstants.ITEM_SLOT_NAMES           = {
	"mainHand",
	"offHand",
	"helm",
	"armour",
	"jewellery",
	"feet",
}

-- This is the value that should be returned from any HandleStartTurn callbacks if the
-- status effect causes the player to miss the next turn
BattleGroundConstants.MISS_TURN						= "miss turn"
BattleGroundConstants.UPDATE_HERO_STATE				= "update hero state"

-- This is a function that can be passed to Grid::Cascade to calculate the
-- drop time for a gem.  Its used in the BattleGround and in a few puzzles.
-- It has a sister function, which will perform the movement of a gem.

function BattleGroundConstants.MakeDropTimeCalculator(gem_step)
	return function (x, y, x2, y2, bottom)
		local dx = math.abs(x - x2) * gem_step
		local dy = math.abs(y - y2) * gem_step
		
		return (y2 - bottom) * BattleGroundConstants.DROP_DELAY + x2 * BattleGroundConstants.DROP_XDELAY + math.sqrt(2 * math.sqrt(dx*dx + dy*dy) / BattleGroundConstants.FALL_ACCELERATION) * 1000 -- 1000ms in a second
	end
end

-- Board layout
function BattleGroundConstants.SetBoardLayoutConstants()
	local offset = GetScreenOffset()
	if offset == 0 then	-- not widescreen
		BattleGroundConstants.LARGE_GEM_SIZE 			= 64			-- 64x64 gems	-- 8x8 grid
		BattleGroundConstants.MEDIUM_GEM_SIZE 			= 51			-- 51x51 gems	-- 10x10 grid
		BattleGroundConstants.SMALL_GEM_SIZE 			= 42			-- 42x42 gems	-- 12x12 grid
		
		BattleGroundConstants.LARGE_GEM_SCALE			= 1.0
		BattleGroundConstants.MEDIUM_GEM_SCALE			= 1.0
		BattleGroundConstants.SMALL_GEM_SCALE			= 1.0
		
		BattleGroundConstants.SPELL_WIDTH				= 200
	else
		BattleGroundConstants.LARGE_GEM_SIZE 			= 108			-- 108x108 gems	-- 8x8 grid
		BattleGroundConstants.MEDIUM_GEM_SIZE 			= 86			-- 86x86 gems	-- 10x10 grid
		BattleGroundConstants.SMALL_GEM_SIZE 			= 72			-- 72x72 gems	-- 12x12 grid
		
		BattleGroundConstants.LARGE_GEM_SCALE			= 768.0 / 1080.0
		BattleGroundConstants.MEDIUM_GEM_SCALE			= 768.0 / 1080.0
		BattleGroundConstants.SMALL_GEM_SCALE			= 768.0 / 1080.0
		
		BattleGroundConstants.SPELL_WIDTH				= 304
	end
end

BattleGroundConstants.SetBoardLayoutConstants()

BattleGroundConstants.PUZZLE_GEM_SIZE 				= 64			-- puzzle gems
BattleGroundConstants.PUZZLE_GEM_SCALE				= 0.8125
BattleGroundConstants.PUZZLE_GEM_STEP 				= 52

BattleGroundConstants.ARENA_LOWERLEFT_X				= 255
BattleGroundConstants.ARENA_LOWERLEFT_Y				= 128

--[[
_G.BattleGroundConstants = BattleGroundConstants

dofile("Assets/Scripts/Battlegrounds/BattleGroundConstantsPlatform.lua")

_G.BattleGroundConstants = nil
--]]
return BattleGroundConstants
