--[[

BattleGroundGameLogic

	Manages the puzzle game.  The host object manages the game states, including:
	  * the state of the board
	  * the state of the heros (players)
	  * who's turn it is

	There is only ONE host object for any game.  That is, for a multiplayer game, only
	one of the players consoles will have a host object.

	The host object typically broadcasts any state changes to all the clients, where
	they are handled by a BattleGroundView.

LAYOUT

	When starting a host, you can specify a board layout.  A board layout is a 2d
	grid like this:

	{
		{" "," "," "," "," "," "," "," "," "," "," "," "},
		{" "," "," "," "," "," "," "," "," "," "," "," "},
		{" "," "," "," "," "," "," "," "," "," "," "," "},
		{" "," "," "," "," "," "," "," "," "," "," "," "},
		{" "," "," "," "," "," "," "," "," "," "," "," "},
		{" "," "," "," "," "," "," "," "," "," "," "," "},
		{" "," "," "," "," "," "," "," "," "," "," "," "},
		{" "," "," "," "," "," "," "," "," "," "," "," "},
		{" "," "," "," "," "," "," "," "," "," "," "," "},
		{"B","X","B","X","B","X"," "," "," "," "," "," "},
		{"G","G","R","G","G","B"," "," "," "," "," "," "},
		{"R","R","G","R","R","B"," "," "," "," "," "," "},
	}

	Any blank locations are filled with random gems using the normal random gem
	selection code.  Each non-empty location is a single character string where
	the character determines the type of gem to automatically place there.  For
	a list of valid character codes, look in Gems.lua.

	Note that board layouts typically go in their own file under the Grids
	sub-directory.  Board layouts placed there are automatically added to the host
	setup screen when hosting a multiplayer game.

GEM DISTRIBUTION:

	To control what gems appear and what frequency they appear at:

	local GEMS = {
		Gems.MakeInitBattleGroundTuple("Red", 25),
		Gems.MakeInitBattleGroundTuple("Green", 25),
		Gems.MakeInitBattleGroundTuple("Blue", 25),
		Gems.MakeInitBattleGroundTuple("Yellow", 25),
		Gems.MakeInitBattleGroundTuple("Skull", 10),
		Gems.MakeInitBattleGroundTuple("Skull5", 5),
		Gems.MakeInitBattleGroundTuple("Gold", 25),
		Gems.MakeInitBattleGroundTuple("Experience", 25),
		Gems.MakeInitBattleGroundTuple("Wildcard", 1),
	}

	The number is a weight.  All weights are normalized.  So the change of a gem
	spawning is its weight over the sum of all gem weights.

GAME OPTIONS:

	The gameOptions parameter passed to Init() can contain:

		turnTimer:			The turn time limit (set to 0 for infite turn limit), in ms.
		turnLimit:			The maximum number of turns for this game (set to 0 for no turn limit).
		
		-- NEW --
		
		gameTimeLimit			The total time the game will last
		turnTimeLimit			The time each turn will last
		gameTurnLimit			The total number of turns the game will last
		modifyHealth			Modify the hit points for each player
		modifyGemProbability	Modify the distribution of gems
		modifyMana				Modify mana pool levels for each player
		replaceItem				Replace the main hand, off hand, or both items for each player

]]

use_safeglobals()

local Grid = require("Battlegrounds/Grid")
local ProbabilityTable = require("Utils/ProbabilityTable")
local Constants = require("Battlegrounds/BattleGroundConstants")

class "BattleGroundGameLogic" (GameObject)

BattleGroundGameLogic.AttributeDescriptions = AttributeDescriptionList()

local LARGE_GEM_STEP 	= Constants.LARGE_GEM_STEP
local LARGE_GEM_SCALE	= Constants.LARGE_GEM_SCALE
local MEDIUM_GEM_STEP 	= Constants.MEDIUM_GEM_STEP
local MEDIUM_GEM_SCALE	= Constants.MEDIUM_GEM_SCALE
local SMALL_GEM_STEP 	= Constants.SMALL_GEM_STEP
local SMALL_GEM_SCALE	= Constants.SMALL_GEM_SCALE

-- Defines the delay for a gem droping that is relative to its height on the board
--  (this delay is the delay that stops all gems in a row from falling at the same time)
local DROP_DELAY		= Constants.DROP_DELAY
local DROP_XDELAY		= Constants.DROP_XDELAY

local WILDCARD_DIST		= Constants.WILDCARD_DIST

local POOLS				= Constants.MANA_POOL_NAMES

local function MutateWithTable(event, data, methodNameOptional)
	local methodName = methodNameOptional
	
	if not methodNameOptional then
		methodName = "MutateEvent" .. event:GetName()
	end

	for k, obj in pairs(data) do
		if obj[methodName] then
			obj[methodName](obj, event)
		end
	end

--	event:do_OnPostMutate()
end

function BattleGroundGameLogic:__init(clid)
	super("BHST")

	self.turnCount = 0
	self.HEROIC_EFFORT_COUNT = Constants.HEROIC_EFFORT_COUNT
	self.PAUSE_BETWEEN_CASCADES = Constants.PAUSE_BETWEEN_CASCADES
	self.cascading = false	-- this will be set to true the very instance a move or spell is approved, and unset when the board is in a steady state
	self.currentMoveID = 1
end

function BattleGroundGameLogic:PreDestroy()
	if leakdetect then
		leakdetect.PreDestroyHook(self)
	end
	scheduler:StopAllThreadsForObject(self)
end

function BattleGroundGameLogic:Init(layout, gems, numClients, gameOptions)
	assert(layout)
	assert(gems)
	assert(numClients)

	if not gameOptions then
		gameOptions = {}
	end

	if not gameOptions.turnTimer then
		gameOptions.turnTimer = 0
	end

	if not gameOptions.turnLimit then
		gameOptions.turnLimit = 0
	end
	
	self.gameOptions = gameOptions

	self:InitGems(gems)
	self:InitLayout(layout)
	self:InitClients(numClients)

	-- keeps track of how much each spell has been cast for each player
	self.spellCastCounts = {}
	
	-- keeps track of how many turns remain on a spell cooldown
	self.spellCoolDowns = {}
end

function BattleGroundGameLogic:SetFinishedHandler(handler)
	self.OnFinished = handler
end

function BattleGroundGameLogic:InitGems(gems)
	assert(gems)

	local distribution = {}

	for i, geminfo in ipairs(gems) do
		distribution[geminfo.name] = geminfo.probability
	end
	
	local modGemProb = self.gameOptions.modifyGemProbability
	if modGemProb then
		for name, probability in pairs(modGemProb.weights) do
			distribution[name] = probability
		end
	end
	
	self.gemProbabilityTable = ProbabilityTable:Construct(distribution)
	self.wildcardProbabilityTable = ProbabilityTable:Construct(WILDCARD_DIST)
end

function BattleGroundGameLogic:FillBoardFromGridLayout(grid)
	assert(grid)

	for y=self.height,1,-1 do
		local y2 = self.height-y+1
		if grid[y] then
			for x=1,self.width do
				local gem = grid[y][x]

				if gem and gem ~= " " then

					if gem == "-" then
						self.grid[x][y2] = self:CreateBlockingGem()
					elseif gem == "=" then
						self.grid[x][y2] = self:CreateSpawningGem()
					elseif gem == "E" then
						self.grid[x][y2] = self:CreateEmptyGem()
					elseif gem ~= " " then
						local gemType = Gems:GetNameFromSymbol(gem)
						assert(gemType, string.format("Grid file '%s' contains an invalid symbol at %d,%d", grid.name, y, x))
						self.grid[x][y2] = self:CreateGem(gemType)
					else
						self.grid[x][y2] = nil
					end
				else
					self.grid[x][y2] = nil
				end
			end
		end
	end
end

function BattleGroundGameLogic:InitLayout(layout)
	assert(layout)

	self.width = layout.width
	self.height = layout.height
	self.grid = Grid:Construct(self.width, self.height)
	self:FillBoardFromGridLayout(layout.grid)

	-- any gems left unspecified by the layout are filled with random gems
	self:FillGems()
end

function BattleGroundGameLogic:CreateRandomGem()
	local name = self.gemProbabilityTable:RandomItem()

	if name == "Wildcard" then
		return self:CreateWildcard()
	end

	local gem = {}

	gem.name = name
	gem.def = Gems[gem.name]

	return gem
end

function BattleGroundGameLogic:CreateGem(colour)
	assert(colour)

	if colour == "Wildcard" then
		return self:CreateWildcard()
	end

	local gem = {}

	gem.name = colour
	gem.def = Gems[colour]
	
	return gem
end

function BattleGroundGameLogic:CreateWildcard()
	local wildcard = {}

	wildcard.name = "Wildcard" .. self.wildcardProbabilityTable:RandomItem()
	wildcard.def = Gems[wildcard.name]

	return wildcard
end

function BattleGroundGameLogic:CreateEmptyGem()
	return {
		name = "Empty",
		def = Gems["Empty"],
	}
end

function BattleGroundGameLogic:CreateBlockingGem()
	return {
		name = "Blocking",
		def = Gems["Blocking"],
	}
end

function BattleGroundGameLogic:ReplaceGem(x, y, color)
	self:ExplodeGem(x, y, false)
	self.grid[x][y] = self:CreateGem(color)

	-- propagate to views
	local newgem = GameEventManager:Construct("BoardSnapshot")

	newgem:PushAttribute("x", x)
	newgem:PushAttribute("y", y)
	newgem:PushAttribute("name", color)
	newgem:SetAttribute("type", "replace")

	GameEventManager:Send(newgem)
end

function BattleGroundGameLogic:TransformGem(x, y, color)
	self.grid[x][y] = self:CreateGem(color)

	-- propagate to views
	local newgem = GameEventManager:Construct("BoardSnapshot")

	newgem:PushAttribute("x", x)
	newgem:PushAttribute("y", y)
	newgem:PushAttribute("name", color)
	newgem:SetAttribute("type", "transform")

	GameEventManager:Send(newgem)
end

function BattleGroundGameLogic:DestroyGem(gem)
	assert(gem)
end

function BattleGroundGameLogic:FindMatches()
	return self.grid:FindMatches()
end

-- Fill any blank spaces on the board with a gem.  This method generates a
-- board with NO matches.  It does not generate any notifications for clients
-- and as such is only suitable for initialisation.
function BattleGroundGameLogic:FillGems(grid)
	if not grid then
		grid = self.grid
	end

	repeat
		for x=1,self.width do
			for y=1,self.height do
				if not grid[x][y] then
					grid[x][y] = self:CreateRandomGem()
				end
			end
		end

		local matches = self:FindMatches(grid)

		if matches then
			-- Pop the middle gems in each match
			for i, match in ipairs(matches) do
				for j=2,#match.gems-1 do
					local geminfo = match.gems[j]

					grid[geminfo.x][geminfo.y] = nil

					self:DestroyGem(geminfo.gem)
				end
			end
		end
	until not matches
end

function BattleGroundGameLogic:InitClients(numClients)
	assert(numClients)

	self.clients = {}
	self.teams = {{},{}}
	self.activeClient = nil
end

function BattleGroundGameLogic:AddClient(team, client)
	LOGF("AddClient: %d", team)

	assert(team)
	assert(client)

	table.insert(self.clients, client);
	table.insert(self.teams[team], client);
	
	client.state = GameObjectManager:Construct("BGHS")
	client.state.host = self
	client.state.client = client
	client:AddChild(client.state)
	client.host = self
	client.bonusTurnCount = 0
	client.turnCount = 0
	client.moveCount = 0
	
	if self.gameOptions.hintArrowRules then
		client.hintArrowRules = self.gameOptions.hintArrowRules[team]
	end
	
	self.spellCastCounts[client] = {}
	self.spellCoolDowns[client] = {}
end

function BattleGroundGameLogic:GetOpposingClient(cl)
	assert(cl)

	for _, info in ipairs(self.clients) do
		if info ~= cl then
			return info
		end
	end

	error("Failure")
end

function BattleGroundGameLogic:SendBoardSnapshot()
	local boardStateEvent = GameEventManager:Construct("BoardSnapshot")

	for x, row in ipairs(self.grid) do
		for y, gem in ipairs(row) do
			boardStateEvent:PushAttribute("x", x)
			boardStateEvent:PushAttribute("y", y)
			boardStateEvent:PushAttribute("name", gem.name)
		end
	end
	
	GameEventManager:Send(boardStateEvent)
end

-- Call when all clients are connected, this will send out initial state messages
-- to all the clients.
function BattleGroundGameLogic:Start(isAmbush)
	LOG("BattleGroundGameLogic:Start()")

	local gameInfoEvent = GameEventManager:Construct("GameInfo")

	gameInfoEvent:SetAttribute("width", self.width)
	gameInfoEvent:SetAttribute("height", self.height)
	gameInfoEvent:SetAttribute("numClients", #self.clients)

	if self.gameOptions.turnTimer ~= 0 then
		gameInfoEvent:SetAttribute("turnTimer", self.gameOptions.turnTimer)
	end
	
	if self.gameOptions.turnLimit ~= 0 then
		gameInfoEvent:SetAttribute("turnLimit", self.gameOptions.turnLimit)
	end

	GameEventManager:Send(gameInfoEvent)

	self:SendBoardSnapshot()

	-- Send information about each client to everyone
	for i, client in ipairs(self.clients) do
		local hero = client.hero
		
		client.state:Initialise(client, hero, self.gameOptions)
		
		if self.gameOptions.addTrait and self.gameOptions.addTrait[i] then
			local data = self.gameOptions.addTrait[i]
			
			for _, name in pairs(data) do
				client.state:AddHeroTrait(client, name)
			end
		end
		
		if self.gameOptions.modifyHealth then
			local data = self.gameOptions.modifyHealth[i]
			if data then
				local startingHealth = data.startingHealth
				if startingHealth then
					client.state.health = startingHealth
					if data.setToMaxHealth then
						client.state.max_health = startingHealth
					end
					
					if client.state.max_health < client.state.health then
						client.state.max_health = client.state.health
					end
				end
				local percentHealth = data.percentHealth
				if percentHealth then
					local health = math.floor(client.state.health * (percentHealth / 100))
					client.state.health = health
					if data.setToMaxHealth then
						client.state.max_health = health
					end
					
					if client.state.max_health < client.state.health then
						client.state.max_health = client.state.health
					end
				end
			end
		end
		
		if self.gameOptions.modifyMana then
			local data = self.gameOptions.modifyMana[i]
			if data then
				for _, pool in pairs(POOLS) do
					if data.startingMana and data.startingMana[pool] then
						client.state[pool] = data.startingMana[pool]
					end
					if data.percentMana and data.percentMana[pool] then
						client.state[pool] = math.floor(client.state[pool] * (data.percentMana[pool] / 100))
					end
					
					local base_max_pool = string.format("base_max_%s", pool)
					if client.state[base_max_pool] and client.state[base_max_pool] < client.state[pool] then
						client.state[pool] = client.state[base_max_pool]
					end
				end
			end
		end
		
		client.state:Broadcast()
	end
	
	-- award starting bonus turns
	self.clients[1].bonusTurnCount = self.clients[1].startingBonusMoves or 0
	self.clients[2].bonusTurnCount = self.clients[2].startingBonusMoves or 0
	
	-- Start the first turn
	local hero1 = self.clients[1].hero
	local hero2 = self.clients[2].hero
	
	local startClientIndex = 2
	if self.clients[1].bonusTurnCount > 0 then
		startClientIndex = 2
	elseif self.clients[2].bonusTurnCount > 0 then
		startClientIndex = 1
	elseif ((hero2.def) and (hero2.def.neverStarts)) and not ((hero1.def) and (hero1.def.neverStarts)) then
		startClientIndex = 2
	elseif ((hero1.def) and (hero1.def.neverStarts)) and not ((hero2.def) and (hero2.def.neverStarts)) then
		startClientIndex = 1
	elseif isAmbush then
		startClientIndex = 1
		self.clients[startClientIndex].state:AddStatusEffect(self.clients[startClientIndex].state, "Stun", "[STUN_AMBUSH]", 5, "[SURPRISED_MONSTER]", "[STUN_NOTIFY_AMBUSH]")
		--self.clients[startClientIndex].bonusTurnCount = self.clients[startClientIndex].bonusTurnCount + 2	-- just in case we want to match the DS and just give extra turns
	elseif hero2:GetStat("agility") > hero1:GetStat("agility") then		-- Also copes for the rare situation where two players have their NeverStartsFirstTurn flag set (maybe in multiplayer tournament?)
		startClientIndex = 1
	elseif hero1:GetStat("agility") == hero2:GetStat("agility") then
		startClientIndex = math.random(1, 2)
	end
	
	self.activeClient = self.clients[startClientIndex]
	self.activeClientID = startClientIndex
	if startClientIndex == 1 then
		self.nextClient = self.clients[2]
	else
		self.nextClient = self.clients[1]
	end
	
	-- we time battles, let this be the starting time!
	self.startGameTime = GetGameTime()
	
	-- NICK: THIS IS VIEW STUFF, IT SHOULD NOT BE HERE !!
	
	if self.gameOptions.conversation then
		open_conversation_menu(
			"Assets/Conversations/" .. self.gameOptions.conversation .. ".xml",
			self.activeClient.hero:GetName(), 
			self.activeClient.hero:GetConversationPortrait(), 
			self.activeClient.hero:GetGenderForConversation(), 
			"", 
			"",
			"", 
			function ()
				self:StartFirstMove()
			end, 
			x or GetScreenWidth()*0.5,
			y or GetScreenHeight()*0.5
		)
	else
		self:StartFirstMove()
	end
end

function BattleGroundGameLogic:StartFirstMove()
	self.activeClient.allowMoves = true
	
	local startNotify = GameEventManager:Construct("BattleStartNotify")
	
	startNotify:SetAttribute("activePlayer", self.nextClient)
	startNotify:SetAttribute("nextPlayer", self.activeClient)
	startNotify:SetAttribute("activePlayerClass", self.nextClient.state.hero:GetAttribute("class"))
	startNotify:SetAttribute("nextPlayerClass", self.activeClient.state.hero:GetAttribute("class"))
	
	GameEventManager:Send(startNotify)
	
	self:StartNextTurn()
end

function BattleGroundGameLogic:ClearGrid()
	for x=1,self.width do
		for y=1,self.height do
			if not self.grid[x][y].def.blocking then
				self.grid[x][y] = nil
			end
		end
	end
end

function BattleGroundGameLogic:SetMoveHasBeenMade()
	self.currentMoveID = self.currentMoveID + 1
	self.activeClient.moveCount = self.activeClient.moveCount + 1
	
	-- NICK: ARGGGGHHHH, THIS DOES NOT WORK FOR MULTIPLAYER!!
	SCREENS.BattleGameMenu:GetWorld():SetCurrentMoveID(self.currentMoveID)
end

-- Host needs to start a new move.  However, before starting a move we always
-- need to check if the board has a valid move on it.  If not, the board is
-- locked and we do a mana drain.
function BattleGroundGameLogic:OnEventProcessStartMove(event)
	local bonus = event:GetAttribute("bonus")
	local client = event:GetAttribute("client")
	
	-- note, in relation to #3397, the current move id is incremented at the end of the 
	-- last move, because otherwise a move timeout can interrupt the next players turn
	-- before it has even started.  (when the battle game waits for accumulated damage
	-- to count down).

	scheduler:AddChildThread(self, function ()
		if self.grid:IsLocked() then
			LOG("Board locked!")
			self.grid:Dump()

			local ManaDrainNotify = GameEventManager:Construct("ManaDrainNotify")
			GameEventManager:Send(ManaDrainNotify)

			-- Drain everyones mana pools
			for _, info in ipairs(self.clients) do
				GameEventManager:Send(GameEventManager:Construct("ManaDrain"), info.state)
			end
			
			-- start popping the gems
			local allgems = table.randomise_entries(self.grid:GetAllGems())
			for i, coord in ipairs(allgems) do
				local gem = self.grid[coord.x][coord.y]
				if not gem.def.blocking then
					self:ExplodeSingleGem(coord.x, coord.y)
					-- dramatic pause!
					scheduler:Sleep(50)
				end
			end
			
			-- It is stunningly remote, but there is a chance that the mana drain could
			-- generate a board with no valid moves again, so keep clearing and filling
			-- until we get something valid
			repeat
				self:ClearGrid()
				self:FillGems()
			until self.grid:FindBestMove()
			
			if IsOpen("BattleGameMenu") then	-- this should help fix a very rare crash
				self:SendBoardSnapshot()
			end
		end

		-- calculate best possible move to show the hint arrow
		local hintX1, hintY1, hintX2, hintY2 = client:FindBestMove(self.grid)
--		LOGF("HintArrow - set coord values: x1=%d, y1=%d, x2=%d, y2=%d", hintX1, hintY1, hintX2, hintY2)
		
		local startMove = GameEventManager:Construct("StartMove")

		startMove:SetAttribute("client", client)
		startMove:SetAttribute("bonus", bonus)
		startMove:SetAttribute("hintX1", hintX1)
		startMove:SetAttribute("hintY1", hintY1)
		startMove:SetAttribute("hintX2", hintX2)
		startMove:SetAttribute("hintY2", hintY2)
		startMove:SetAttribute("moveID", self.currentMoveID)

		GameEventManager:Send(startMove)
		
		if self.gameOptions.turnTimer ~= 0 then
			local moveTimeout = GameEventManager:Construct("MoveTimeout")
			moveTimeout:SetAttribute("moveID", self.currentMoveID)
			GameEventManager:SendDelayed(moveTimeout, self, GetGameTime() + self.gameOptions.turnTimer)
		end
	end, "OnEventProcessStartMove")
end

function BattleGroundGameLogic:AddTurnTimerToTurn(duration, frenzy)
	assert(type(duration) == "number")
	assert(self.gameOptions.turnTimer == 0, "what should frenzy do, ... if turn times are limited already?!")
	
	local moveTimeout = GameEventManager:Construct("MoveTimeout")
	moveTimeout:SetAttribute("moveID", self.currentMoveID)
	moveTimeout:SetAttribute("frenzy", (frenzy and 1) or 0)
	GameEventManager:SendDelayed(moveTimeout, self, GetGameTime() + duration)
end

function BattleGroundGameLogic:OnEventMoveTimeout(event)
	-- check this isnt a timeout for a move that has since ended of natural causes
	if event:GetAttribute("moveID") == self.currentMoveID then
		if event:GetAttribute("frenzy") == 1 then
			self.activeClient.state:ResetFrenzied()
		end
		if self.cascading then
			-- player has made their move, actually, nothing needs to be done here
		else
			self:CheckEndOfTurn()
		end
	end
end

function BattleGroundGameLogic:CanUseSpells(client)
	return MutateValueWithStatusEffects("CanUseSpells", client.state.statusEffects, true)
end

function BattleGroundGameLogic:IsSpellUseCountExceeded(client, spellName, spellInfo)
	return spellInfo.castLimit and self.spellCastCounts[client][spellName] and self.spellCastCounts[client][spellName] >= spellInfo.castLimit
end

-- A spell request has been made from the client.
function BattleGroundGameLogic:OnEventSpellRequest(event)
	assert(event)

	local client = event:GetAttribute("client")

	if client ~= self.activeClient then
		return self:RejectSpell(event, false, "Rejecting move because its from an inactive client")
	end

	if not client.allowMoves then
		return self:RejectSpell(event, false, "Rejecting move because the client needs to wait for cascades to finish")
	end

	local targetX = event:GetAttribute("targetX")
	local targetY = event:GetAttribute("targetY")

	if targetX ~= 0 and targetY ~= 0 then -- 0,0 means "not set"
		if targetX < 1 or targetX > self.width or targetY < 1 or targetY > self.height then
			return self:RejectSpell(event, false, "Rejecting spell since the target is out of bounds")
		end

		if not self.grid[targetX][targetY] then
			return self:RejectSpell(event, false, "Rejecting spell since the target nil")
		end
	end
	
	-- some status effects can BLOCK spell casting
	if not MutateValueWithStatusEffects("CanUseSpells", self.activeClient.state.statusEffects, true) then
		return self:RejectSpell(event, false, "Rejecting spell since the player is stunned!")
	end
	
	local spells = self.activeClient.state.spells
	local spellIndex = event:GetAttribute("index")
	
	-- index could be malicously out of range
	if not (spells[spellIndex]) then
		return self:RejectSpell(event, false, "Rejecting spell since the index is out of range")
	end
	
	local spellName = spells[spellIndex]
	local spellInfo = SPELLS[spellName]

	-- if spell requires targetting, make sure a target was specified
	if spellInfo.requiresTarget then
		if targetX == 0 and targetY == 0 then
			return self:RejectSpell(event, false, "Rejecting spell since it requires a target and none was specified")
		end
	end

	-- check mana requirements of spell
	for _, pool in pairs(POOLS) do
		if spellInfo.cost[pool] and spellInfo.cost[pool] > client.state[pool] then
			return self:RejectSpell(event, true, string.format("Rejecting spell because the client lacks the mana to cast it: %d vs %d [%s]", spellInfo.cost[pool], client.state[pool], pool))
		end
	end
	
	-- check the cooldown requirements
	local target = self:GetOpposingClient(self.activeClient)
	
	if spellInfo.cooldown and self.spellCoolDowns[target][spellName] and self.spellCoolDowns[target][spellName] > 0 then
		return self:RejectSpell(event, true, string.format("Rejecting spell because the spell is cooling down! (turns until spell active: %d)", self.spellCoolDowns[target][spellName]))
	end
	
	-- check the use count requirement.  some spells have a limited number of uses per battle
	if self:IsSpellUseCountExceeded(client, spellName, spellInfo) then
		return self:RejectSpell(event, false, string.format("Rejecting spell because the client has cast it too often already: %d (the limit is %d)", self.spellCastCounts[client][spellName], spellInfo.castLimit))
	end
	
	-- spell will be cast, ensure that no move timeout will interrupt the game flow (#3354)
	self:SetMoveHasBeenMade()

	local params = {}

	params.source = self.activeClient
	params.target = self:GetOpposingClient(self.activeClient)
	params.grid = self.grid
	params.host = self
	params.targetX = targetX
	params.targetY = targetY
	params.spellName = spellName

	-- TESTING HOOK
	if _G.app.TestingHooks and _G.app.TestingHooks.OnSpellCast then
		_G.app.TestingHooks.OnSpellCast(spellName)
	end

	-- some status effects can CHANGE the effect of an spell
	local spellOverride = MutateValueWithStatusEffects("GetSpellActivationOverride", self.activeClient.state.statusEffects, nil)
	local request_id = event:GetAttribute("request_id")
	
	-- all other requirements met, does the spell have any specific requirements of its own?
	if spellInfo.ArePrerequisitesMet and not spellInfo:ArePrerequisitesMet(params) then
		local notify = GameEventManager:Construct("SpellInvalidNotify")
		notify:SetAttribute("spell", spellOverride or spellName)
		notify:SetAttribute("index", spellIndex)
		notify:SetAttribute("request_id", request_id)
		notify:SetAttribute("client", client)

		-- nothing else needed
		GameEventManager:Send(notify)
		return
	end
	
	-- reset the cooldown counter
	local cool_down = spellInfo.cooldown	--spellInfo.cooldownForAI?
	self.spellCoolDowns[target][spellName] = cool_down or 0
	
	if cool_down and cool_down > 0 then
		local notify = GameEventManager:Construct("SpellCooldownNotify")
		notify:SetAttribute("cooldown", cool_down)
		notify:SetAttribute("spell", spellName)
		notify:SetAttribute("client", client)
		GameEventManager:Send(notify)
	end
	
	-- spell resistance
	local spellPenetration = params.source.state:GetSpellPenetration()
	if params.target.state:IsSpellResisted(spellPenetration) then
		-- stop this client making any more moves
		client.allowMoves = false
		scheduler:AddChildThread(self, function ()
			local notify = GameEventManager:Construct("SpellResistedNotify")
			notify:SetAttribute("spell", spellOverride or spellName)
			notify:SetAttribute("index", spellIndex)
			notify:SetAttribute("request_id", request_id)
			notify:SetAttribute("client", client)
			GameEventManager:Send(notify)
			
			-- Pause for a bit, a delay between spell resist and turn end is nice
			scheduler:Sleep(2 * Constants.PAUSE_AFTER_SPELL)
			
			self:CheckEndOfTurn()
		end)
		
		return
	end
	
	-- we keep a counter of how many times each spell has been cast
	self.spellCastCounts[client][spellName] = (self.spellCastCounts[client][spellName] or 0) + 1
	
	-- spell will be cast, enter cascade mode
	self.cascading = true

	-- stop this client making any more moves
	client.allowMoves = false

	scheduler:AddChildThread(self, function ()
		-- send notification to clients
		local notify = GameEventManager:Construct("SpellNotify")

		notify:SetAttribute("spell", spellOverride or spellName)
		notify:SetAttribute("index", spellIndex)
		notify:SetAttribute("request_id", request_id)
		notify:SetAttribute("client", client)
		
		params.notify = notify

		-- apply the spell costs
		local spellCostModifier = self.activeClient.state:GetActualSpellCostModifier()
		local totalManaUsed = 0
		
		for _, pool in pairs(POOLS) do
			if spellInfo.cost[pool] then
				local drain = GameEventManager:Construct("DamageMana")
				local cost = math.floor(spellInfo.cost[pool] * (spellCostModifier / 100))
				
				totalManaUsed = totalManaUsed + cost

				drain:SetAttribute("amount", cost)
				drain:SetAttribute("source", params.source)
				drain:SetAttribute("pool", pool)

				GameEventManager:Send(drain, client.state)
			end
		end
		
		-- update stats
		local stats = client.state.stats
		
		stats.total_mana_used = stats.total_mana_used + totalManaUsed
		stats.spell_use[spellIndex] = stats.spell_use[spellIndex] + 1

		if not spellOverride then
			local spell = spellInfo:Apply(params, notify)

			-- Send this after the Spell:Apply method since that method might attack
			-- extra parameters to the event
			GameEventManager:Send(notify)
			
			-- Some status effects response to spell activations
			MutateValueWithStatusEffects("HandleSpellActivation", self.activeClient.state.statusEffects, notify)

			if spell then
				spell()
			end
		else
			-- This is for the status effect "The Lads" which is turns the enemy players
			-- next attack into a spell that destroys all the green gems on the boards 
			-- (the green gems representing the "lads")
			
			local overrideSpellInfo = SPELLS[spellOverride]
			local spell = overrideSpellInfo:Apply(params, notify)

			-- Send this after the Spell:Apply method since that method might attack
			-- extra parameters to the event
			GameEventManager:Send(notify)

			if spell then
				spell()
			end
		end

		if params.requireMainHandStrike and self.activeClient.state.mainHandItem then
			self:ProcessItemActivation(self.activeClient, Items.MAIN_HAND, nil --[[ no request id]], params.requiresCascade, params.requiresCheckMatches, true)
		elseif params.requireOffHandStrike and self.activeClient.state.offHandItem then
			self:ProcessItemActivation(self.activeClient, Items.OFF_HAND, nil --[[ no request id]], params.requiresCascade, params.requiresCheckMatches, true)
		elseif params.requireCascade then
			-- any spell that has the chance to replace an empty gem with a real gem should require 
			-- a cascade (NOT JUST A CHECK MATCHES!) because you could create a gem with empty space
			-- under it, and the gem should fall on top of another gem before proceeding to check
			-- for matches or ending the players turn
			
			local checkForMatches = GameEventManager:Construct("CheckForMatches")

			checkForMatches:SetAttribute("client", client) -- this is who any cascading matches would effect
			checkForMatches:SetAttribute("count", 1)

			scheduler:Sleep(Constants.PAUSE_AFTER_SPELL)
			
			self:Cascade(checkForMatches)
		elseif params.requireCheckMatches then
			-- any spell that modifies the state of the board should set this flag so that the logic
			-- checks for matches after the spell is cast.  See above for "requireCascade" and the
			-- exception to this rule
			
			local checkForMatches = GameEventManager:Construct("CheckForMatches")

			checkForMatches:SetAttribute("client", client) -- this is who any cascading matches would effect
			checkForMatches:SetAttribute("count", 1)

			GameEventManager:SendDelayed(checkForMatches, self, GetGameTime() + Constants.PAUSE_AFTER_SPELL) -- delay slighly, so we can see the effect the spell had before blowing crap up
		else
			-- Pause for a bit, allows damage to register on players, and a delay between spell effects and turn end is nice too
			scheduler:Sleep(Constants.PAUSE_AFTER_SPELL)
			
			self.cascading = false -- turns out we never needed to cascade from this spell
			self:CheckEndOfTurn()
		end
	end)
end

function BattleGroundGameLogic:CanUseWeapons(client)
	return MutateValueWithStatusEffects("CanUseItems", client.state.statusEffects, true)
end

-- The player is wanting to activate an item
function BattleGroundGameLogic:OnEventItemActivationRequest(event)
	assert(event)

	local client = event:GetAttribute("client")

	if client ~= self.activeClient then
		return self:RejectItemActivation(event, nil, false, "Rejecting item activation because its from an inactive client")
	end

	if not client.allowMoves then
		return self:RejectItemActivation(event, nil, false, "Rejecting item activation because the client needs to wait for cascades to finish")
	end
	
	-- NOTE: it is quite possible that some items will require a target location on the
	-- board to cast into (like some spells).

	local itemIndex = event:GetAttribute("index")
	local itemObject
	
	if itemIndex == Items.MAIN_HAND then
		itemObject = self.activeClient.state.mainHandItem
	elseif itemIndex == Items.OFF_HAND then
		itemObject = self.activeClient.state.offHandItem
	else
		return self:RejectItemActivation(event, nil, false, "Rejecting item activation you can only activate items in the players hands")
	end
	
	if not client.state:IsStealthed() and itemObject:IsUsableOnlyInStealth() then
		return self:RejectItemActivation(event, nil, false, "Rejecting item activation because it requires stealth and the client is not in stealth mode")
	end
	
	if not itemObject then
		return self:RejectItemActivation(event, nil, false, string.format("Rejecting item activation because there is no such item at index %d", itemIndex))
	end
	
	local itemBase = itemObject.baseDef

	-- check that the caster has sufficient action points
	local itemCostModifier = client.state:GetStealthModeItemCostModifier(itemObject)
	if itemObject:GetActionPower(client.state.hero) * itemCostModifier > client.state.power then
		return self:RejectItemActivation(event, itemObject, false, string.format("Rejecting item activation because the client lacks the action points to cast it: has %d needs %d", client.state.power, itemObject:GetActionPower() * itemCostModifier))
	end

	-- some status effects can BLOCK item activation
	if not MutateValueWithStatusEffects("CanUseItems", self.activeClient.state.statusEffects, true) then
		return self:RejectItemActivation(event, nil, false, "Rejecting item activation since the player is stunned!")
	end
	
	return self:ProcessItemActivation(client, itemIndex, event:GetAttribute("request_id"))
end

function BattleGroundGameLogic:ProcessItemActivation(client, itemIndex, request_id, requiresCascade, requiresCheckMatches, ignoreCost)
	assert(client)
	assert(itemIndex)
	
	local itemObject
	
	if itemIndex == Items.MAIN_HAND then
		itemObject = self.activeClient.state.mainHandItem
	elseif itemIndex == Items.OFF_HAND then
		itemObject = self.activeClient.state.offHandItem
	end
	
	local itemBase = assert(itemObject).baseDef
	
	-- item will be used, ensure that no move timeout will interrupt the game flow (#3354)
	self:SetMoveHasBeenMade()
	
	-- spell will be cast, enter cascade mode
	self.cascading = true
	
	-- stop this client making any more moves
	client.allowMoves = false

	local params = {}

	params.caster = self.activeClient
	params.source = self.activeClient
	params.target = self:GetOpposingClient(self.activeClient)
	params.grid = self.grid
	params.host = self
	params.targetX = targetX
	params.targetY = targetY
	params.requiresCascade = requiresCascade
	params.requiresCheckMatches = requiresCheckMatches
	
	local damageAmount = 0
	local defenceAmount = 0
	local defenceTurns = 0
	local critical = false
	
	if itemObject:IsWeapon() then
		damageAmount = itemObject:GetDamageBase()
		damageAmount = damageAmount + client.state:GetWeaponBonusDamage(damageAmount)
			critical = client.state:IsWeaponCritical()
	elseif itemObject:IsShield() then
		defenceAmount, defenceTurns = itemObject:GetActiveDefenseBonus(client.state.hero)
		critical = client.state:IsShieldCritical()
	end
	
	if critical then
		damageAmount = damageAmount * 2
		defenceAmount = defenceAmount * 2
	end
	
	-- some status effects perform actions when an item is used
	for statusEffect in pairs(params.source.state.statusEffects) do
		statusEffect:HandleMyItemActivation(itemObject)
	end
	
	for statusEffect in pairs(params.target.state.statusEffects) do
		statusEffect:HandleItemActivation(itemObject)
	end
	
	-- some status effects can CHANGE the effect of an item
	local itemSpell = MutateValueWithStatusEffects("GetItemActivationOverride", self.activeClient.state.statusEffects, nil)
	
	-- send notification to clients
	local notify = GameEventManager:Construct("ItemActivationNotify")
	notify:SetAttribute("item", itemIndex)
	notify:SetAttribute("index", itemIndex)
	notify:SetAttribute("request_id", request_id or 0)
	notify:SetAttribute("client", client)
	notify:SetAttribute("damage_amount", damageAmount )
	notify:SetAttribute("defence_amount", defenceAmount )
	notify:SetAttribute("critical", critical)
	notify:SetAttribute("damage_missed", toboolean(damageAmount == 0 and itemObject:IsWeapon()))
	notify:SetAttribute("portrait", client.hero:GetAttribute("portrait"))
	notify:SetAttribute("backdrop", LEVELS[client.hero:GetCurrentLocationLevelName()].combatBanner)
	
	local value = itemObject:GetActivationValue()
	if itemObject:IsHealthPotion() then
		notify:SetAttribute("health_amount", value)
	elseif itemObject:IsManaPotion() then
		notify:SetAttribute("mana_amount", value)
	elseif itemObject:IsManaAccelerant() then
		notify:SetAttribute("manamatch_amount", value)
	elseif itemObject:IsPoison() then
		notify:SetAttribute("damageturn_amount", value)
	end
	
	GameEventManager:Send(notify)
	
	-- TESTING HOOK
	if _G.app.TestingHooks and _G.app.TestingHooks.OnItemUsed then
		_G.app.TestingHooks.OnItemUsed(itemObject)
	end
	
	scheduler:AddChildThread(self, function ()
	
		scheduler:Sleep(Constants.PAUSE_FOR_ITEM_EFFECT)
	
		-- remove action points from the caster
		local itemCostModifier = client.state:GetStealthModeItemCostModifier(itemObject)
		local itemCost = itemObject:GetActionPower(client.state.hero) * itemCostModifier
		
		if itemCost > 0 and not ignoreCost then
			local drain = GameEventManager:Construct("DamageMana")

			drain:SetAttribute("amount", itemCost)
			drain:SetAttribute("source", params.source)
			drain:SetAttribute("pool", "power")

			GameEventManager:Send(drain, client.state)
		end

		params.source.state:OnItemActivated(itemObject)

		if not itemSpell then
			-- Apply any magical effects that might need to go off when striking the player - this 
			-- could return immediately, or take a while (the magical effect might cast a spell)

			itemObject:ApplyMagicalEffects("OnStrike", params)

			if itemObject:IsWeapon() then
				-- Apply damage from the weapon to the player (only if this is a weapon)
				local attack = GameEventManager:Construct("DamageHealth")
				
				attack:SetAttribute("amount", damageAmount)
				attack:SetAttribute("source", params.source)
				
				GameEventManager:Send(attack, params.target.state)
				
				-- status effects have handlers for damage that need to be called
				MutateWithTable(attack, self.activeClient.state.statusEffects, "HandleDamageGiven")
				MutateWithTable(attack, params.target.state.statusEffects, "HandleDamageRecieved")
					
				params.source.state:OnWeaponUsed(itemObject, damageAmount)
			elseif itemObject:IsShield() then
				-- Shields do a temporary buff to defense rating
				-- which amounts to a status effect on the player
				local applied_turns = defenceTurns * 2
				self.activeClient.state:AddStatusEffect(self.activeClient.state, "ModifyDefense", itemObject:GetName(), applied_turns, itemObject:GetName(), 0, defenceAmount)
			end
		else
			-- This is for the status effect "The Lads" which is turns the enemy players
			-- next attack into a spell that destroys all the green gems on the boards 
			-- (the green gems representing the "lads")
					
			local notify = GameEventManager:Construct("SpellNotify")

			notify:SetAttribute("spell", itemSpell)
			notify:SetAttribute("index", -1)
			notify:SetAttribute("request_id", -1)
			notify:SetAttribute("client", client)

			GameEventManager:Send(notify)

			local spellInfo = SPELLS[itemSpell]
			local spell = spellInfo:Apply(params)

			if spell then
				spell()
			end
		end
		
		if params.requireCascade then
			-- any spell that has the chance to replace an empty gem with a real gem should require 
			-- a cascade (NOT JUST A CHECK MATCHES!) because you could create a gem with empty space
			-- under it, and the gem should fall on top of another gem before proceeding to check
			-- for matches or ending the players turn
			
			local checkForMatches = GameEventManager:Construct("CheckForMatches")

			checkForMatches:SetAttribute("client", client) -- this is who any cascading matches would effect
			checkForMatches:SetAttribute("count", 1)

			scheduler:Sleep(Constants.PAUSE_AFTER_ITEM)
			
			self:Cascade(checkForMatches)
		elseif params.requireCheckMatches then
			local checkForMatches = GameEventManager:Construct("CheckForMatches")

			checkForMatches:SetAttribute("client", client) -- this is who any cascading matches would effect
			checkForMatches:SetAttribute("count", 1)

			GameEventManager:SendDelayed(checkForMatches, self, GetGameTime() + Constants.PAUSE_AFTER_ITEM) -- delay slighly, so we can see the effect the spell had before blowing crap up
		else
			-- Pause for a bit, allows damage to register on players, and a delay between spell effects and turn end is nice too
			scheduler:Sleep(Constants.PAUSE_AFTER_ITEM)
			
			self.cascading = false -- turns out we never needed to cascade from this spell
			self:CheckEndOfTurn()
		end
	end)
end

function BattleGroundGameLogic:RejectSpell(event, pending, reason)
	assert(event)
	assert(type(pending) ~= "nil")
	assert(reason)

	local client = event:GetAttribute("client")
	local reject = GameEventManager:Construct("SpellRejected")

	reject:SetAttribute("request_id", event:GetAttribute("request_id"))
	reject:SetAttribute("client", client)
	reject:SetAttribute("pending", pending)

	GameEventManager:Send(reject)

	local spells = client.state.spells
	local index = event:GetAttribute("index")
	
	if spells[index] then
		LOGF("Host has rejected a spell request %s(%d) from client %s because: %s", spells[index], event:GetAttribute("request_id"), tostring(client), reason)
	else
		LOGF("Host has rejected a spell request %d(%d) from client %s because: %s", index, event:GetAttribute("request_id"), tostring(client), reason)
	end
end

function BattleGroundGameLogic:RejectItemActivation(event, item, pending, reason)
	assert(event)
	assert(type(pending) ~= "nil")
	assert(reason)

	local client = event:GetAttribute("client")
	local reject = GameEventManager:Construct("ItemActivationRejected")

	reject:SetAttribute("request_id", event:GetAttribute("request_id"))
	reject:SetAttribute("client", client)
	reject:SetAttribute("pending", pending)

	GameEventManager:Send(reject)

	if item then
		LOGF("Host has rejected an item activation request %s(%d) from client %s because: %s", item:GetName(), event:GetAttribute("request_id"), tostring(client), reason)
	else
		LOGF("Host has rejected an item activation request (%d) from client %s because: %s", event:GetAttribute("request_id"), tostring(client), reason)
	end
end

function BattleGroundGameLogic:ApplyStatusEffect(from, target, name, prettyName, turns, ...)
	return target.state:AddStatusEffect(from, name, prettyName, turns, ...)
end

function BattleGroundGameLogic:RejectMove(event, reason)
	assert(event)
	assert(reason)

	local client = event:GetAttribute("client")
	local swapRejected = GameEventManager:Construct("SwapRejected")
	swapRejected:SetAttribute("source_x", event:GetAttribute("source_x"))
	swapRejected:SetAttribute("source_y", event:GetAttribute("source_y"))
	swapRejected:SetAttribute("target_x", event:GetAttribute("target_x"))
	swapRejected:SetAttribute("target_y", event:GetAttribute("target_y"))
	swapRejected:SetAttribute("request_id", event:GetAttribute("request_id"))
	swapRejected:SetAttribute("client", client)

	GameEventManager:Send(swapRejected)

	LOGF("Rejected move %d,%d <-> %d,%d from %s because: %s",
		event:GetAttribute("source_x"),
		event:GetAttribute("source_y"),
		event:GetAttribute("target_x"),
		event:GetAttribute("target_y"),
		tostring(client),
		tostring(reason)
	)
end

function BattleGroundGameLogic:OnEventSwapRequest(event)
	LOG("BattleGroundGameLogic:OnEventSwapRequest()")

	local client = event:GetAttribute("client")

	if client ~= self.activeClient then
		return self:RejectMove(event, "Rejecting move because its from an inactive client")
	end

	if not client.allowMoves then
		return self:RejectMove(event, "Rejecting move because the client needs to wait for cascades to finish")
	end

	local sx, sy = event:GetAttribute("source_x"), event:GetAttribute("source_y")
	local tx, ty = event:GetAttribute("target_x"), event:GetAttribute("target_y")

	local valid, reason = self.grid:IsValidMove(sx, sy, tx, ty)

	if not valid then
		return self:RejectMove(event, reason)
	end
	
	-- move will be made, ensure that no move timeout will interrupt the game flow (#3354)
	self:SetMoveHasBeenMade()

	-- move will be made, enter cascade mode
	self.cascading = true

	-- stop this client making any more moves
	client.allowMoves = false
	
	SoundSystem.Play("pq2audio/board/gemmove")
	self.grid:SwapGems(sx, sy, tx, ty)

	-- and let the clients know
	local swapNotify = GameEventManager:Construct("SwapNotify")

	swapNotify:SetAttribute("source_x", sx)
	swapNotify:SetAttribute("source_y", sy)
	swapNotify:SetAttribute("target_x", tx)
	swapNotify:SetAttribute("target_y", ty)
	swapNotify:SetAttribute("request_id", event:GetAttribute("request_id"))
	swapNotify:SetAttribute("client", event:GetAttribute("client"))

	GameEventManager:Send(swapNotify)

	-- wait for half a second (this is how long it takes the views to swap to gems) before cascading
	local checkForMatches = GameEventManager:Construct("CheckForMatches")

	checkForMatches:SetAttribute("client", event:GetAttribute("client")) -- this is who any cascading matches would effect
	checkForMatches:SetAttribute("count", 1)

	GameEventManager:SendDelayed(checkForMatches, self, GetGameTime() + Constants.PAUSE_AFTER_MOVE)
	
	-- A move was made, this might effect some status effects
	for effect in pairs(self.activeClient.state.statusEffects) do
		effect:HandleMove(self.activeClient)
	end
end

function BattleGroundGameLogic:AwardBonusMoveToActivePlayer()
	self.activeClient.bonusMove = true
	self.activeClient.state:OnBonusTurn()
end

function BattleGroundGameLogic:AddBonusTurnsToActivePlayer(count)
	self.activeClient.bonusTurnCount = self.activeClient.bonusTurnCount + assert(count)
end

function BattleGroundGameLogic:SkipActivePlayerTurn()
end

function BattleGroundGameLogic:NotifyMatch(match)
	assert(match)
	-- notify the views, they respond to this event by spawning MASSIVE AMOUNTS OF PARTICLES
	local notify = GameEventManager:Construct("MatchNotify")

	notify:SetAttribute("client", self.activeClient)
	notify:SetAttribute("count", match.count)
	notify:SetAttribute("score", match.score)
	notify:SetAttribute("bonus", match.bonus)
	notify:SetAttribute("name", match.name)
	
	for i, info in ipairs(match.gems) do
		notify:PushAttribute("x", info.x)
		notify:PushAttribute("y", info.y)
		notify:PushAttribute("gem", info.gem.name)
	end

	GameEventManager:Send(notify)	
end

function BattleGroundGameLogic:NotifySkullMatch(match, target)
	assert(match)
	assert(target)
	-- notify the views, they respond to this event by spawning MASSIVE AMOUNTS OF PARTICLES
	local notify = GameEventManager:Construct("SkullMatchNotify")

	notify:SetAttribute("client", self.activeClient)
	notify:SetAttribute("count", match.count)
	notify:SetAttribute("score", match.score)
	notify:SetAttribute("bonus", match.bonus)
	notify:SetAttribute("name", match.name)
	notify:SetAttribute("ignored", toboolean((target:IsStealthed() or self.activeClient.state:IsStealthed()) and Gems[match.name].givesDamage))
	
	for i, info in ipairs(match.gems) do
		notify:PushAttribute("x", info.x)
		notify:PushAttribute("y", info.y)
		notify:PushAttribute("gem", info.gem.name)
	end

	GameEventManager:Send(notify)	
end


function BattleGroundGameLogic:ApplyMatchEffects(match)
	assert(match)

	--print( "BattleGroundGameLogic:ApplyMatchEffects\nmatch:" )
	--table.pretty_print( match )	
	
	--print( "BattleGroundGameLogic:ApplyMatchEffects()" )
	--table.pretty_print( match )
	
	local opposingClient = self:GetOpposingClient(self.activeClient)

	-- give the items and status effects the chance to respond to this match,
	-- BEFORE we generate any other game actions from it

	-- We mutate the match from active client perspective
	self.activeClient.state:MutateMatch(match, opposingClient.state)
	
	--print( "BattleGroundGameLogic:ApplyMatchEffects - active:MutateMatch\nmatch:" )
	--table.pretty_print( match )
	
	-- We mutate the match from opposing client perspective
	opposingClient.state:MutateOpponentMatch(match, opposingClient.state)
	
	--print( "BattleGroundGameLogic:ApplyMatchEffects - opposing:MutateOpponentMatch\nmatch:" )
	--table.pretty_print( match )
	
	-- [SC] TODO: move this into the above mutation methods, its a client.state method
	local params = {}
	
	params.caster = self.activeClient
	params.source = self.activeClient
	params.target = opposingClient
	params.grid = self.grid
	params.host = self
	
	self.activeClient.state:ApplyMagicalItemEffects("OnMatch", params, match)
	
	--print( "BattleGroundGameLogic:ApplyMatchEffects - active:ApplyMagicalItemEffects\nmatch:" )
	--table.pretty_print( match )
	
	if params.awardBonusTurn then
		self:AwardBonusMoveToActivePlayer()
	end

	if match.count >= Constants.BONUS_TURN_COUNT then
		self:AwardBonusMoveToActivePlayer()
	end

	-- [SC] TODO: score effectively determines the base magnitude of the mana/damage
	-- and can be mutated by the clients who apply their own states respectively, this
	-- effectively deprecates the call to MutateWithTable(...) if we handle items and 
	-- status effects within client.state:MutateMatch/MutateOpponentMatch (which we dont)
	if match.bonus == 0 then
		match.bonus = 1 -- ensure we dont wipe our score value
	end
	
	local score = match.bonus * match.score

	if Gems[match.name].givesMana then
		
		self:NotifyMatch(match)
		
		local restoreMana = GameEventManager:Construct("RestoreMana")
		if Gems[match.name].givesManaProxy then
			restoreMana:SetAttribute("pool", string.lower(Gems[match.name].givesManaProxy))
		else
			restoreMana:SetAttribute("pool", string.lower(match.name))
		end
		restoreMana:SetAttribute("amount", score)
		restoreMana:SetAttribute("healer", self.activeClient)

		-- mutate the restore
		MutateWithTable(restoreMana, self.activeClient.hero:GetPassiveItems())
		MutateWithTable(restoreMana, self.activeClient.state.statusEffects)

		GameEventManager:Send(restoreMana, self.activeClient.state)
	elseif Gems[match.name].givesDamage then
		--local critical = self.activeClient.state:IsSkullCritical()
		
		-- sends notification to the views
		self:NotifySkullMatch(match, opposingClient.state)
		
		--print( "BattleGroundGameLogic:ApplyMatchEffects - NotifySkullMatch\nmatch:" )
		--table.pretty_print( match )
		
		-- players in stealth do not take damage, players in stealth cannot do skull match damage
		if not self.activeClient.state:IsStealthed() then
			local damageHealth = GameEventManager:Construct("DamageHealth")
			
			-- [SC] TODO: criticals are a flat multiplier to score based on some chance, however
			-- the score is mutated by hero stats, active/passive items, status effects.
			-- At which point the critical multiplier is applied can be determined within the 
			-- client.state as part of the mutation pipeline that applies to a match
			--if critical then
			--	score = 2 * score
			--end
			
			damageHealth:SetAttribute("amount", score)
			damageHealth:SetAttribute("source", self.activeClient)
			damageHealth:SetAttribute("reason", "match")

			-- mutate the damage
			MutateWithTable(damageHealth, self.activeClient.hero:GetPassiveItems())
			MutateWithTable(damageHealth, self.activeClient.state.statusEffects)
			
			-- get your attack
			--local attack = self.activeClient.state:GetActualAttackRating()
			
			-- mutate on the defence value
			--damageHealth:SetAttribute("amount", damageHealth:GetAttribute("amount") + attack)
			
			-- should be done after the attack rating is added
			MutateWithTable(damageHealth, self.activeClient.state.statusEffects, "HandleDamageGiven")
			MutateWithTable(damageHealth, opposingClient.state.statusEffects, "HandleDamageRecieved")
			
			-- OnStruck needs to be called on any items on the target player
			opposingClient.state:ApplyItemMagicalEffects("OnStruck", {
				target = opposingClient.state.client,
				source = self.activeClient,
			})
			
			GameEventManager:Send(damageHealth, opposingClient.state)
		end
	elseif match.name == "Power" then
		self:NotifyMatch(match)
		local givePower = GameEventManager:Construct("GivePower")

		givePower:SetAttribute("amount", score)

		-- mutate the restore
		MutateWithTable(givePower, self.activeClient.hero:GetPassiveItems())
		MutateWithTable(givePower, self.activeClient.state.statusEffects)

		GameEventManager:Send(givePower, self.activeClient.state)
	end
end

function BattleGroundGameLogic:ExplodeMatches(matches)
	-- We explode matches with a 2 pass method: 
	-- 1st pass is to recursively explode each of the gems and their consequentially exploded gems
	-- 2nd pass is to apply the effects of each of those gems and their consequentially exploded gems
	-- This way we ensure that gems with destroy effects (skull5) arent applied to gems
	-- that are already exploded
	
	--print( "BattleGroundGameLogic:ExplodeMatches - matches" )
	--table.pretty_print( matches )
	
	local exploded_gems = {}
	
	local event = GameEventManager:Construct("MatchesNotify")
	for i, match in ipairs(matches) do
		for j, gem in ipairs(match.gems) do
			LOGF("Host exploding %d,%d => %s", gem.x, gem.y, self.grid[gem.x][gem.y].name)
			
			local board_gem = self.grid[gem.x][gem.y]
			if board_gem.name ~= "Empty" then
				event:SetAttribute("total_num", event:GetAttribute("total_num") + 1)
				if board_gem.name == "Red" then
					event:SetAttribute("red_matched", true)
				elseif board_gem.name == "Yellow" then
					event:SetAttribute("yellow_matched", true)
				elseif board_gem.name == "Green" then
					event:SetAttribute("green_matched", true)
				elseif board_gem.name == "Blue" then
					event:SetAttribute("blue_matched", true)
				elseif board_gem.name == "Black" then
					event:SetAttribute("black_matched", true)
				elseif board_gem.name == "Skull" then
					event:SetAttribute("skull_matched", true)
				elseif board_gem.name == "Skull5" then
					event:SetAttribute("skull5_matched", true)
				elseif board_gem.name == "Power" then
					event:SetAttribute("power_num", event:GetAttribute("power_num") + 1)
				elseif board_gem.name == "Wildcard2" or
				       board_gem.name == "Wildcard3" or
				       board_gem.name == "Wildcard4" or
				       board_gem.name == "Wildcard5" or
				       board_gem.name == "Wildcard6" or
				       board_gem.name == "Wildcard7" then
					event:SetAttribute("wildcard_matched", true)
				end
				
				-- ExplodeGem returns a table of any consequentially exploded gems from the exploded gem
				exploded_gems = table.join( exploded_gems, self:ExplodeGem( gem.x, gem.y, true ) )
			end
		end
	end
	
	--print( "exploded_gems" )
	--table.pretty_print( exploded_gems )
	
	-- The only source of duplicates might be gems that are part of the original match, they will be accounted for
	-- in the match calculation so we avoid applying single gem effects for these gems
	local function is_matched_gem(x, y)
		for i, match in ipairs(matches) do
			for j, gem in ipairs(match.gems) do
				if x == gem.x and y == gem.y then
					return true
				end
			end
		end
		return false
	end
	
	for i, exploded in ipairs( exploded_gems ) do
		if not is_matched_gem( exploded.x, exploded.y ) then
			self:ApplyEffectForSingleGem( exploded.gem, exploded.x, exploded.y )
		end
	end
	
	GameEventManager:Send(event)
	for i, match in ipairs(matches) do
		-- Give the active client the effects of these gems
		self:ApplyMatchEffects(match)

		if match.count >= 5 then
			-- place a wildcard in the middle gem
			local middle = match.gems[math.floor(#match.gems / 2) + 1]

			self.grid[middle.x][middle.y] = self:CreateWildcard()

			-- notify views
			local snapshot = GameEventManager:Construct("BoardSnapshot")

			snapshot:PushAttribute("x", middle.x)
			snapshot:PushAttribute("y", middle.y)
			snapshot:PushAttribute("name", self.grid[middle.x][middle.y].name)
			snapshot:SetAttribute("type", "match")

			GameEventManager:Send(snapshot)
		end
	end
end

function BattleGroundGameLogic:ApplyHeroicEffort(match)
	assert(match)
	
	-- Give 100 xp
	local giveExperience = GameEventManager:Construct("GiveBattleExperience")
	
	giveExperience:SetAttribute("amount", Constants.HEROIC_EFFORT_XP)
--	giveExperience:SetAttribute("heroic", true)
	
	GameEventManager:Send(giveExperience, self.activeClient.state)
	
	-- Place a wildcard in middle of match
	local middle = match.gems[math.floor(#match.gems / 2) + 1]
	
	self.grid[middle.x][middle.y] = self:CreateWildcard()
	
	-- notify views
	local snapshot = GameEventManager:Construct("BoardSnapshot")
	
	snapshot:PushAttribute("x", middle.x)
	snapshot:PushAttribute("y", middle.y)
	snapshot:PushAttribute("name", self.grid[middle.x][middle.y].name)
	snapshot:SetAttribute("type", "match")
	
	GameEventManager:Send(snapshot)
	
	-- Award an extra turn
	self:AwardBonusMoveToActivePlayer()
	
	-- Send a notify to the views
	--  (not required, the views get cascade notifies which contain the cascade index)
end

function BattleGroundGameLogic:SendBonusMove()
	-- Tell the client it can move
	local startMove = GameEventManager:Construct("ProcessStartMove")

	startMove:SetAttribute("client", self.activeClient)
	startMove:SetAttribute("host", self)
	startMove:SetAttribute("bonus", true)

	GameEventManager:Send(startMove, self)
end

function BattleGroundGameLogic:OnEventCheckForMatches(event)
	-- A cascade has just finished
	local matches = self:FindMatches()

	if matches then
		self:ExplodeMatches(matches)
		
		if event:GetAttribute("count") == self.HEROIC_EFFORT_COUNT then
			self:ApplyHeroicEffort(matches[1])
		end
		
		-- need to do a cascade, after this cascade we need to recheck the board
		local checkForMatches = GameEventManager:Construct("CheckForMatches")

		checkForMatches:SetAttribute("client", event:GetAttribute("client"))
		checkForMatches:SetAttribute("count", event:GetAttribute("count") + 1)

		self:Cascade(checkForMatches)
	else
		self.grid:Dump()
		
		self.cascading = false  -- no more matches, cascade has finished
		self:CheckEndOfTurn()
	end
end

function BattleGroundGameLogic:CheckEndOfTurnInternal()
	-- check victory conditions
	if self:CheckVictoryConditions() then
		return
	end

	-- turn over, someone else can move
	if self.activeClient.bonusMove then
		self.activeClient.bonusMove = false
		self.activeClient.allowMoves = true
		self:SendBonusMove()
	else
		self:StartNextTurn()
	end
end

function BattleGroundGameLogic:CheckEndOfTurn()
	-- THIS IS WHERE FRENZY IS IMPLEMENTED
	-- If the player is frenzied then their turn does not end until the frenzy
	-- has worn off.  So if this is the case, we do no further processing except
	-- to send a bonus move
	if self.activeClient.state:IsFrenzied() then
		self.activeClient.allowMoves = true
		if not self.activeClient.frenzyTriggered then
			self.activeClient.frenzyTriggered = true
			self.activeClient.frenzyTurnCount = 1
		else
			self.activeClient.frenzyTurnCount = self.activeClient.frenzyTurnCount + 1
		end
		self:SendBonusMove()
		return 
	end
	
	if self.activeClient.frenzyTriggered then
		self.activeClient.bonusMove = false				-- once frenzy is disabled, you lose any bonus turns
		self.activeClient.frenzyTriggered = false		-- reset immediately
		if self.activeClient.state:ResetFrenzied(self.activeClient.frenzyTurnCount) ~= "endturn" then
			return
		end
	end
	

	-- At this stage we apply all damage that the hero accumulated over
	-- the duration of this turn.  Firstly we send notifications to the
	-- view as the view want to show all manner of FX containers to make
	-- it BLINDINGLY obivous to the player what is happening ...
	
	local damagePause = 0
	local damageTaken = false
	local damageAmounts = {}
	
	for i, client in ipairs(self.clients) do
		local original, adjusted = client.state:ApplyDamagePreProcessing()
		
		if original ~= adjusted then
			damagePause = math.max(100 + (original - adjusted) * 25 + 100, damagePause)
		end
		
		if original > 0 or adjusted > 0 then
			damageTaken = true
			damageAmounts[client] = {original=original, adjusted=adjusted}
		end
	end
	
	if damageTaken then
		-- If a client had damage recieved then the following calls will
		-- actually apply the damage to the client.  AFTER a three second
		-- delay, which should give the views enough time to show their 
		-- fancy pants effects.
		
		scheduler:AddChildThread(self, function ()
			scheduler:Sleep(damagePause)
			
			for i, client in ipairs(self.clients) do
				if damageAmounts[client] then
					client.state:ApplyDamagePostProcessing(damageAmounts[client].original, damageAmounts[client].adjusted)
				end
				
				-- invariants
				assert(client.state.damage == 0)
			end
			
			self:CheckEndOfTurnInternal()
		end)
	else	
		self:CheckEndOfTurnInternal()
		
		-- invariants
		for i, client in ipairs(self.clients) do
			assert(client.state.damage == 0)
		end
	end

end

function BattleGroundGameLogic:IsNPCFight()
	local npc = false
	
	for i, player in ipairs(self.teams[1]) do
		if player:GetAttribute("npc") then
			npc = true
		end
	end
	
	for i, player in ipairs(self.teams[2]) do
		if player:GetAttribute("npc") then
			npc = true
		end
	end
	
	return npc
end

function BattleGroundGameLogic:FinishGame(winner, loser)
	assert(winner)
	assert(loser)

	-- we should send the final players state so that for single player
	-- games we can add on XP and Gold (for example).
	
	-- player could end on a bonus turns streak, update this before quitting
	self.activeClient.state.stats.best_bonus_moves_streak = math.max(self.activeClient.state.stats.best_bonus_moves_streak, self.activeClient.state.stats.num_bonus_moves_this_turn)
	self.activeClient.state.stats.num_bonus_moves_this_turn = 1
	
	-- send stats for every player
	self.teams[1][1].state:SendStats(self.teams[2][1])
	self.teams[2][1].state:SendStats(self.teams[1][1])
	
	local endNotify = GameEventManager:Construct("BattleEndNotify")
	endNotify:SetAttribute("winningTeam", winner)
	endNotify:SetAttribute("winningPlayer", self.teams[winner][1])
	endNotify:SetAttribute("losingPlayer", self.teams[loser][1])
	GameEventManager:Send(endNotify)

	-- the game has ended, send one final message
	local finished = GameEventManager:Construct("BattleFinished")

	finished:SetAttribute("winningTeam", winner)
	finished:SetAttribute("winningPlayer", self.teams[winner][1])
	finished:SetAttribute("winningClass", self.teams[winner][1].state.hero:GetAttribute("class"))
	finished:SetAttribute("winningGender", self.teams[winner][1].state.hero:GetGenderForStats())
	finished:SetAttribute("losingPlayer", self.teams[loser][1])
	finished:SetAttribute("losingClass", self.teams[loser][1].state.hero:GetAttribute("class"))
	finished:SetAttribute("losingGender", self.teams[loser][1].state.hero:GetGenderForStats())
	finished:SetAttribute("score1", 131)
	finished:SetAttribute("score2", 121)
	finished:SetAttribute("health1", self.teams[1][1].state.health)
	finished:SetAttribute("health2", self.teams[2][1].state.health)
	finished:SetAttribute('end_xp1', self.teams[1][1].state.xp)
	finished:SetAttribute('end_xp2', self.teams[2][1].state.xp)
	finished:SetAttribute('turn_count', self.turnCount)
	finished:SetAttribute('turn_count1', self.teams[1][1].turnCount)
	finished:SetAttribute('turn_count2', self.teams[2][1].turnCount)
	finished:SetAttribute('move_count1', self.teams[1][1].moveCount)
	finished:SetAttribute('move_count2', self.teams[2][1].moveCount)
	finished:SetAttribute('npc_fight', self:IsNPCFight())
	finished:SetAttribute('play_time', math.floor((GetGameTime() - self.startGameTime) / 1000))
	
	GameEventManager:Send(finished)

	-- This event is sent to the game logic 2.5 seconds into the
	-- future.  It triggers the game termination (BattleCleanupEnd)
	GameEventManager:SendDelayed(GameEventManager:Construct("BattleCleanupStart"), self, GetGameTime() + 2500)

	return true
end

function BattleGroundGameLogic:OnEventBattleCleanupStart()
	-- broadcast the end of the game to everyone in the game
	local event = GameEventManager:Construct("BattleCleanupEnd")
	GameEventManager:Send(event)
end

function BattleGroundGameLogic:CheckVictoryConditions()
	local team1AliveCount = 0
	local team2AliveCount = 0

	for i, player in ipairs(self.teams[1]) do
		if not player.state:IsDead() then
			team1AliveCount = team1AliveCount + 1
		end
	end

	for i, player in ipairs(self.teams[2]) do
		if not player.state:IsDead() then
			team2AliveCount = team2AliveCount + 1
		end
	end

	if team1AliveCount == 0 or team2AliveCount == 0 then
		local winner = 0
		local loser = 0

		if team1AliveCount > 0 then
			winner = 1
			loser = 2
			
			SCREENS[self.clients[2]]:FadeDown()
		else
			winner = 2
			loser = 1
			
			SCREENS[self.clients[1]]:FadeDown()
		end
	
		self:FinishGame(winner, loser)
		
		return true
	end
end

function BattleGroundGameLogic:OnEventStartNextTurnAfterMissingMove()
	self:StartNextTurn()
end

function BattleGroundGameLogic:GetCurrentTurnCount()
	return self.turnCount
end

function BattleGroundGameLogic:StartNextTurn()
	if self.gameOptions.turnLimit > 0 and self.turnCount >= self.gameOptions.turnLimit then
		-------------------------------------------------------------------------------------
		-- This is how it used to work, its probably the way it *should* work
		-------------------------------------------------------------------------------------
		
		--local winner = 0
		--local loser = 0
		--
		--if self.teams[1][1].state.health > self.teams[2][1].state.health then
		--	winner = 1
		--	loser = 2
		--	
		--	SCREENS[self.clients[2]]:FadeDown()
		--else
		--	winner = 2
		--	loser = 1
		--	
		--	SCREENS[self.clients[1]]:FadeDown()
		--end
		
		--self:FinishGame(winner, loser)
		
		-- However, we only use the turn timer in challenge fights, and the designers actually
		-- wanted it to fail if you hit the turn threshold.  So, we just finish the game and
		-- make the player the loser (#2886)
		
		self:FinishGame(2, 1)

		return
	end

	-- note, we dont send the final EndTurn message because it would be recieved on
	-- at the view after the client objects have been destroyed.

	if self.turnCount > 0 then
		local endTurn = GameEventManager:Construct("EndTurn")
		assert(self.activeClient)
		assert(self.turnCount)
		endTurn:SetAttribute("client", self.activeClient)
		endTurn:SetAttribute("turn", self.turnCount)

		GameEventManager:Send(endTurn)
	end
	
	for effect in pairs(self.activeClient.state.statusEffects) do
		effect:HandleEndTurn(self.activeClient)
	end
	
	-- Update status effects at the end of each turn
	for i, client in ipairs(self.clients) do
		client.state:UpdateStatusEffects()
	end
	
	if self.activeClient.bonusTurnCount > 0 then
		self.activeClient.bonusTurnCount = self.activeClient.bonusTurnCount - 1
	elseif self.nextClient.bonusTurnCount < 0 then
		self.nextClient.bonusTurnCount = self.nextClient.bonusTurnCount + 1
	else
		self.activeClientID = self.activeClientID + 1

		if self.activeClientID > #self.clients then
			self.activeClientID = 1
		end
	end

	self.turnCount = self.turnCount + 1
	self.activeClient.allowMoves = false
	self.activeClient.state.stats.best_bonus_moves_streak = math.max(self.activeClient.state.stats.best_bonus_moves_streak, self.activeClient.state.stats.num_bonus_moves_this_turn)
	self.activeClient.state.stats.num_bonus_moves_this_turn = 1
	self.activeClient = self.clients[self.activeClientID]
	self.activeClient.turnCount = self.activeClient.turnCount + 1
	
	self.nextClientID = self.activeClientID + 1
	if self.nextClientID > #self.clients then
		self.nextClientID = 1
	end
	self.nextClient = self.clients[self.nextClientID]
	
	-- Apply any status effect bonuses and penalties
	local missTurn = false
	local missTurnReason
	
	for effect in pairs(self.activeClient.state.statusEffects) do
		local ret, reason = effect:HandleStartTurn(self.activeClient)
		
		if ret == Constants.MISS_TURN then
			missTurn = true
			missTurnReason = reason
		elseif ret == Constants.UPDATE_HERO_STATE then
			self.activeClient.state:Broadcast()
		end
	end
	
	-- The rest of the processessing is defered until the next frame after
	-- we will have recieved any DOT damage messages from the HandleStartTurn
	-- above.  This is to fix #3238.
	
	local continueStartNextTurn = GameEventManager:Construct("ContinueStartNextTurn")
	
	continueStartNextTurn:SetAttribute("missTurn", toboolean(missTurn))
	continueStartNextTurn:SetAttribute("missTurnReason", missTurnReason or "")
	
	GameEventManager:Send(continueStartNextTurn, self)
end

function BattleGroundGameLogic:OnEventContinueStartNextTurn(event)
	local missTurn = event:GetAttribute("missTurn")
	local missTurnReason = event:GetAttribute("missTurnReason")
	
	if missTurnReason == "" then
		missTurnReason = nil
	end

	if self:CheckVictoryConditions() then
		-- player died and game is won, stop anything else from happening (#1182)
		return
	end
	
	LOGF("Active Client = %s(%d)", tostring(self.activeClient), self.activeClientID)
	self.activeClient.allowMoves = true
	self.activeClient.frenzyTriggered = false
	self.activeClient.frenzyTurnCount = 0

	-- Note, just because we tell the client its their turn doesnt mean that they are
	-- allowed to move.  They could still have this turn skipped by a status effect 
	-- (like "Stun").  When the player does miss a turn, the UI on the view acts as if
	-- the player was about to start a move before showing a "miss turn" FX container
	-- and switching turns again immediately.
	
	local startTurn = GameEventManager:Construct("StartTurn")

	startTurn:SetAttribute("client", self.activeClient)
	startTurn:SetAttribute("opponent", self.nextClient)
	startTurn:SetAttribute("turn", self.turnCount)
	startTurn:SetAttribute("missTurn", toboolean(missTurn))

	GameEventManager:Send(startTurn)
	
	-- stealth effects hook start of turn
	self.activeClient.state:OnStartTurn()
	
	if not missTurn then
		-- Tell the client it can move
		local startMove = GameEventManager:Construct("ProcessStartMove")
		startMove:SetAttribute("client", self.activeClient)
		GameEventManager:Send(startMove, self)
		
		-- Decrement the cool downs, this is only done when the player actually gets a move.
		for spell, value in pairs(self.spellCoolDowns[self.activeClient]) do
			if value > 0 then
				self.spellCoolDowns[self.activeClient][spell] = value - 1
				
				local notify = GameEventManager:Construct("SpellCooldownNotify")
				notify:SetAttribute("cooldown", value - 1)
				notify:SetAttribute("spell", spell)
				local target = self:GetOpposingClient(self.activeClient)
				notify:SetAttribute("client", target)
				GameEventManager:Send(notify)
			end
		end
	else
		LOG("Player is missing this turn!")
		
		-- never tell the client they can move, send a miss turn notify, then
		-- send a delayed event to trigger the start of the next turn.
		SCREENS.BattleGameMenu:GetWorld().allowMoves = false -- THIS MAKES ME ANGRY
		local missMove = GameEventManager:Construct("MissMove")
		missMove:SetAttribute("client", self.activeClient)
		missMove:SetAttribute("reason", missTurnReason or "[BLANK]")
		GameEventManager:Send(missMove)
		self.activeClient.allowMoves = false
		GameEventManager:SendDelayed(GameEventManager:Construct("StartNextTurnAfterMissingMove"), self, GetGameTime() + Constants.PAUSE_AFTER_MISS_MOVE)
	end
end

function BattleGroundGameLogic:OnEventExitStealth(event)
	-- stub, no longer needed
end

function BattleGroundGameLogic:ApplyEffectForSingleGem(gem, x, y)
	assert(gem)
	assert(x)
	assert(y)

	--print( string.format( "BattleGroundGameLogic:ApplyEffectForSingleGem(%s, %d, %d)", tostring( gem.name ), x, y ) )
	
	if gem.name ~= "Empty" then
		local match = {}

		match.name = gem.name
		match.count = 1
		match.score = gem.def.score or 1
		match.bonus = 1
		match.gems = {{ x=x, y=y, gem=gem }}

		self:ApplyMatchEffects(match)
	end

end

function BattleGroundGameLogic:ExplodeSingleGem(x, y)
	self.grid[x][y] = self:CreateEmptyGem()

	-- notify clients
	local explodeGem = GameEventManager:Construct("ExplodeGem")

	explodeGem:SetAttribute("x", x)
	explodeGem:SetAttribute("y", y)

	GameEventManager:Send(explodeGem)
end

-- does NOT apply effects from exploding the gem, does NOT apply effects from exploding consequentially exploded gems,
-- it does return a table of any consequentially exploded gems so you can then iterate and apply effects if needed.
-- duplicates are suppressed by replacing exploded gems with empty gems, and ignoring instructions to explode empty gems.
function BattleGroundGameLogic:ExplodeGem(gem_x, gem_y, do_consequential_explosions)
	assert(gem_x)
	assert(gem_y)

	local consequential_explosions = {}
	
	if gem_x > 0 and gem_x <= self.width and gem_y > 0 and gem_y <= self.height then
		local exploding_gem = table.clone( self.grid[gem_x][gem_y] )
		local gem_name = exploding_gem.name
		
		if gem_name ~= "Empty" then
			print( string.format( "BattleGroundGameLogic:ExplodeGem( %d, %d, %s )", gem_x, gem_y, tostring( do_consequential_explosions ) ) )
			-- only non-empty gems can actually explode, this also prevents duplicate explosions
			-- on gems that have become empty during this explosion sequence.
			self:ExplodeSingleGem( gem_x, gem_y )
			
			-- if this is an explosive skull, destroy the gems around this one as well
			if do_consequential_explosions and gem_name == "Skull5" then
				for explode_x = gem_x - 1, gem_x + 1 do
					for explode_y = gem_y - 1, gem_y + 1 do
						if ( explode_x ~= gem_x or explode_y ~= gem_y ) and explode_x > 0 and explode_x <= self.width and explode_y > 0 and explode_y <= self.height then
							local explode_gem = table.clone( self.grid[explode_x][explode_y] )
							table.insert( consequential_explosions, { x = explode_x, y = explode_y, gem = explode_gem } )
							consequential_explosions = table.join( consequential_explosions, self:ExplodeGem( explode_x, explode_y, do_consequential_explosions ) )
						end
					end
				end
			end
		end
	end
	
	--print( "consequential_explosions" )
	--table.pretty_print( consequential_explosions )
	
	return consequential_explosions
end

function BattleGroundGameLogic:Cascade(nextEvent)
	assert(nextEvent)

	LOG("BattleGroundGameLogic:Cascade()")

	-- for each empty spot, determine the gem that will fall into it
	local cascadeEvent = GameEventManager:Construct("CascadeNotify")
	local delay = 0
	local gem_step = SCREENS.BattleGameMenu:GetWorld().gem_step -- HMMMMMMMM: FIX ME?

	delay = self.grid:Cascade(cascadeEvent, Constants.MakeDropTimeCalculator(gem_step))

	local left = cascadeEvent:GetAttribute("left")
	local right = cascadeEvent:GetAttribute("right")
	local top = cascadeEvent:GetAttribute("top")
	local bottom = cascadeEvent:GetAttribute("bottom")

	local function ExpandRegion(x, y)
		if y > top then
			top = y
		end

		if y < bottom then
			bottom = y
		end

		if x > right then
			right = x
		end

		if x < left then
			left = x
		end
	end

	for x=1,self.width do
		for y=1,self.height do
			if self.grid[x][y].name == "Empty" then
				local x2, y2 = self.grid:FindNextGem(x, y, function (gem) return gem.def.blocking end)

				if not x2 then
					self:DestroyGem(self.grid[x][y])

					self.grid[x][y] = self:CreateRandomGem()

					cascadeEvent:PushAttribute("newx", x)
					cascadeEvent:PushAttribute("newy", y)
					cascadeEvent:PushAttribute("newname", self.grid[x][y].name)

					ExpandRegion(x, y)

					local dx = 0
					local dy = math.abs(y - self.height+1) * gem_step
					local localDelay = (y - bottom) * DROP_DELAY + x * DROP_XDELAY + math.sqrt(2 * math.sqrt(dx*dx + dy*dy) / Constants.FALL_ACCELERATION) * 1000

					delay = math.max(delay, localDelay)
				end
			end
		end
	end

	cascadeEvent:SetAttribute("left", left)
	cascadeEvent:SetAttribute("right", right)
	cascadeEvent:SetAttribute("top", top)
	cascadeEvent:SetAttribute("bottom", bottom)
	cascadeEvent:SetAttribute("count", nextEvent:GetAttribute("count"))

	GameEventManager:Send(cascadeEvent)

	GameEventManager:SendDelayed(nextEvent, self, GetGameTime() + delay + self.PAUSE_BETWEEN_CASCADES)

	LOG("Board, post cascade:")

	self.grid:Dump()
end

-- REMOVE FROM RETAIL (TODO)

function BattleGroundGameLogic:CheatCycleGem(x, y, to)
	LOGF("CHEATING: cycled gem %d,%d from %s to %s", x, y, self.grid[x][y].def.name, to)
	self:TransformGem(x, y, to)
end

function BattleGroundGameLogic:CheatKillTeam(id)
	for i, player in ipairs(self.teams[id]) do
		player.state:DamageHealth(player, player.state.health, "Hahaha")
	end
end

function BattleGroundGameLogic:CheatForceEndOfGameCheck()
	-- This is a good way to force the end of game check that is
	-- delayed by a few frames
	local checkForMatches = GameEventManager:Construct("CheckForMatches")

	checkForMatches:SetAttribute("client", client) -- this is who any cascading matches would effect
	checkForMatches:SetAttribute("count", 1)

	GameEventManager:SendDelayed(checkForMatches, self, GetGameTime() + Constants.PAUSE_AFTER_SPELL) -- delay slighly, so we can see the effect the spell had before blowing crap up
end

function BattleGroundGameLogic:CheatBattleWin()
	self:CheatKillTeam(1)
	self:CheatForceEndOfGameCheck()
end

function BattleGroundGameLogic:CheatBattleLose()
	self:CheatKillTeam(2)
	self:CheatForceEndOfGameCheck()
end

function BattleGroundGameLogic:CheatInitLayouts()
	self.boardLayouts = {}
	self.boardLayoutIndex = 0

	local gridsList = import("Grids/GridsList")
	for _, name in ipairs(gridsList.playable) do
		table.insert(self.boardLayouts, require("Grids/" .. name))
	end
	for _, name in ipairs(gridsList.testing) do
		table.insert(self.boardLayouts, require("Grids/" .. name))
	end
end

function BattleGroundGameLogic:CheatChangeBoard(grid)
	LOGF("CHEATING: board reset to %s", grid.name)
	SCREENS.BattleGameMenu:set_text("str_name", grid.name)	-- REMOVE LATER

	for x=1,self.width do
		for y=1,self.height do
			self.grid[x][y] = self:CreateEmptyGem()
		end
	end

	self.grid:Dump()
	self:FillBoardFromGridLayout(grid)
	self:FillGems()
	self:SendBoardSnapshot()
end

function BattleGroundGameLogic:CheatNextBoardLayout()
	if not self.boardLayouts then
		self:CheatInitLayouts()
	end

	self.boardLayoutIndex = self.boardLayoutIndex + 1

	if self.boardLayoutIndex > #self.boardLayouts then
		self.boardLayoutIndex = 1
	end

	self:CheatChangeBoard(self.boardLayouts[self.boardLayoutIndex])
end

function BattleGroundGameLogic:CheatPreviousBoardLayout()
	if not self.boardLayouts then
		self:CheatInitLayouts()
	end

	self.boardLayoutIndex = self.boardLayoutIndex - 1

	if self.boardLayoutIndex < 1 then
		self.boardLayoutIndex = #self.boardLayouts
	end

	self:CheatChangeBoard(self.boardLayouts[self.boardLayoutIndex])
end

function BattleGroundGameLogic:CheatManaGain(id, amount_in)
	local amount = amount_in or 20
	
	for i, player in ipairs(self.teams[id]) do
		player.state.actual_max_red = player.state.hero_invariant_stats.derived.mana.maximum.red + amount
		player.state:RestoreMana(player, "red", player.state.actual_max_red)
		player.state.actual_max_green = player.state.hero_invariant_stats.derived.mana.maximum.green + amount
		player.state:RestoreMana(player, "green", player.state.actual_max_green)
		player.state.actual_max_blue = player.state.hero_invariant_stats.derived.mana.maximum.blue + amount
		player.state:RestoreMana(player, "blue", player.state.actual_max_blue)
		player.state.actual_max_yellow = player.state.hero_invariant_stats.derived.mana.maximum.yellow + amount
		player.state:RestoreMana(player, "yellow", player.state.actual_max_yellow)
		player.state.actual_max_black = player.state.hero_invariant_stats.derived.mana.maximum.black + amount
		player.state:RestoreMana(player, "black", player.state.actual_max_black)
		player.state.power = player.state.power + 1
	end
end

function BattleGroundGameLogic:CheatManaAdd(id, amount)
	assert(amount)
	for i, player in ipairs(self.teams[id]) do
		player.state:RestoreMana(player, "red", amount)
		player.state:RestoreMana(player, "green", amount)
		player.state:RestoreMana(player, "blue", amount)
		player.state:RestoreMana(player, "yellow", amount)
		player.state:RestoreMana(player, "black", amount)
	end
end

function BattleGroundGameLogic:CheatLeftManaGain(amount)
	self:CheatManaGain(1, amount)
end

function BattleGroundGameLogic:CheatRightManaGain()
	self:CheatManaGain(2, amount)
end

function BattleGroundGameLogic:CheatLeftManaAdd(amount)
	self:CheatManaAdd(1, amount)
end

function BattleGroundGameLogic:CheatHealthGain(id)
	for i, player in ipairs(self.teams[id]) do
		player.state.max_health = player.state.max_health + 20
		player.state:RestoreHealth(player, player.state.max_health)
	end
end

function BattleGroundGameLogic:CheatLeftHealthGain()
	self:CheatHealthGain(1)
end

function BattleGroundGameLogic:CheatRightHealthGain()
	self:CheatHealthGain(2)
end

function BattleGroundGameLogic:RandomiseColourOfManaGems(numGems)
	local function SetupNewManaGem(gem)
		local colours = {"Red", "Green", "Blue", "Yellow", "Black"}
		for j, name in ipairs(colours) do
			if name == gem.name then
				table.remove(colours, j)
				break
			end
		end
		local colour = table.random(colours)
		local x = gem.x
		local y = gem.y
		self:TransformGem(x, y, colour)
	end
	
	local manaGems = self.grid:GetAllGemsByAttribute("givesMana")
	if #manaGems <= numGems then
		repeat
			for i=1, #manaGems do
				SetupNewManaGem(manaGems[i])
			end
		until not self:FindMatches(self.grid)
--	else	-- TODO: work out how to deal with matches without changing more gems than is wanted
--		repeat
--			for i=1, numGems do
--				local index = math.random(1, #manaGems)
--				local gem = table.remove(manaGems, index)
--				SetupNewManaGem(gem)
--			end
--		until not self:FindMatches(self.grid)
	end
end

function BattleGroundGameLogic:AddOwnerToGem(x, y, owner)
	assert(x)
	assert(y)
	assert(owner)
	
	if x >= 1 and x <= self.width and y >= 1 and y <= self.height then
		self.grid[x][y].owner = owner
	end
end

function BattleGroundGameLogic:RemoveOwnerFromGem(x, y)
	assert(x)
	assert(y)
	
	if x >= 1 and x <= self.width and y >= 1 and y <= self.height then
		self.grid[x][y].owner = nil
	end
end

return ExportClass("BattleGroundGameLogic")
