
local GEMDROP_CHANCE = 0.15

local ItemGenerator = require("Items.Generator")
local ItemConfiguration = require("Items.ItemConfiguration")
local LootTables = require("LootTables.LootTables")

local function SelectMonsterLevel(def, heroLevel)
	assert(def)
	assert(heroLevel)
	
	local monsterLevel = math.random(heroLevel - def.minLevelBelowPlayer, heroLevel + def.maxLevelAbovePlayer)
	
	if monsterLevel < def.minLevel then
		monsterLevel = def.minLevel
	elseif monsterLevel > def.maxLevel then
		monsterLevel = def.maxLevel
	end
	
	return monsterLevel
end

local function CreateItem(itemInfo, level, slot)
	if itemInfo.uniqueType then
		return CreateUniqueItem(itemInfo.uniqueType, level)
	elseif itemInfo.tradeType then
		return CreateTradeItem(itemInfo.tradeType, itemInfo.stackSize or 1)
	else
		local params = {
			slot = slot,
			level = level,
			baseType = itemInfo.baseType,
			race = itemInfo.race,
			rarity = itemInfo.rarity,
		}
		return ItemGenerator.RandomEquippableItem(params)
	end
end

function RefreshMoveIntelligence(monsterHero, heroDifficulty)
	assert(monsterHero)
	assert(heroDifficulty)
	
	local move_intelligence
	local def = monsterHero.def
	local monsterMoveIntelligence = def.moveIntelligence or 0
	
	if heroDifficulty == 0 then -- Easy
		move_intelligence = (monsterMoveIntelligence * 5) + 5
	elseif heroDifficulty == 1 then -- Normal
		move_intelligence = monsterMoveIntelligence
	elseif heroDifficulty == 2 then -- Hard
		move_intelligence = 0
	end
	
	monsterHero:SetAttribute("move_intelligence", move_intelligence)
end

function DestroyMonsterHero(monster)
	GameObjectManager:Destroy(monster)
end

function CreateMonsterWithoutReferenceHero(monster, heroLevel, difficulty, nodeID, die)


	return mathex.scoped_dice(die or cDice(), function ()
		local def = require("Monsters." .. monster)
		
		assert(def.hitPointsPerLevel)
		assert(def.texttag)
		assert(def.portrait)
		
		local heroClassID
		
		if def.class then
			heroClassID = require(string.format("Hero.Default%s", def.class)).classID
		else
			heroClassID = "HERO"
		end
		
		local hero = GameObjectManager:ConstructLocal(heroClassID)
		local level = SelectMonsterLevel(def, heroLevel)
		
		-- Increase Stats based on level
		local mStrength = def.strength or 1
		local mAgility = def.agility or 1
		local mStamina = def.stamina or 1
		local mIntelligence = def.intelligence or 1
		local mMorale = def.morale or 1
		if (level > def.minLevel) then
			mStrength     = mStrength + math.floor((level-def.minLevel)*mStrength/(def.minLevel+2))
			mAgility      = mAgility + math.floor((level-def.minLevel)*mAgility/(def.minLevel+2))
			mStamina      = mStamina + math.floor((level-def.minLevel)*mStamina/(def.minLevel+2))
			mIntelligence = mIntelligence + math.floor((level-def.minLevel)*mIntelligence/(def.minLevel+2))
			mMorale       = mMorale + math.floor((level-def.minLevel)*mMorale/(def.minLevel+2))
		end
		
		hero:SetAttribute("name", string.format("[%s_NAME]", def.texttag))
		hero:SetAttribute("level", level)
		hero:SetAttribute("portrait", def.portrait)
		
		hero:SetAttribute("strength", mStrength)
		hero:SetAttribute("agility", mAgility)
		hero:SetAttribute("stamina", mStamina)
		hero:SetAttribute("intelligence",mIntelligence)
		hero:SetAttribute("morale", mMorale)
		hero:SetAttribute("move_intelligence", def.moveIntelligence or 0)
		hero.def = def
		
		-- Set the move intelligence based on the referenceHero difficulty
		RefreshMoveIntelligence(hero, difficulty)
		
		if nodeID then
			hero:SetAttribute("currentLocation", string.format("R%s", string.sub(nodeID, 1, 3)))
			hero:SetAttribute("currentNode", nodeID)
		end
		
		if def.class then
			hero:SetAttribute("class", def.class)
		end
		
		if def.primaryWeaponSlot then
			if (not def.primaryWeaponSlot.chance) or (math.random() <= def.primaryWeaponSlot.chance) then
				local item = CreateItem(def.primaryWeaponSlot, level, "main")
				if item then
					hero:SetAttribute("mainHandItem", item)
					hero:AddChild(item)
				end
			end
		end
		
		local mainHandItem = hero:GetAttribute("mainHandItem")
		if mainHandItem and mainHandItem:IsTwoHanded() then
			hero:SetAttribute("offHandItem", mainHandItem)
		else
			if def.secondaryWeaponSlot and not twoHanded then
				if (not def.secondaryWeaponSlot.chance) or (math.random() <= def.secondaryWeaponSlot.chance) then 
					local item = CreateItem(def.secondaryWeaponSlot, level, "off")
					if item then
						hero:SetAttribute("offHandItem", item)
						hero:AddChild(item)
					end
				end
			end
		end
		
		if def.armourSlot then
			if (not def.armourSlot.chance) or (math.random() <= def.armourSlot.chance) then 
				local item = CreateItem(def.armourSlot, level, "armour")
				if item then
					hero:SetAttribute("armourItem", item)
					hero:AddChild(item)
				end
			end
		end
		
		if def.helmSlot then
			if (not def.helmSlot.chance) or (math.random() <= def.helmSlot.chance) then 
				local item = CreateItem(def.helmSlot, level, "helm")
				if item then
					hero:SetAttribute("helmItem", item)
					hero:AddChild(item)
				end
			end
		end
		
		if def.bootsSlot then
			if (not def.bootsSlot.chance) or (math.random() <= def.bootsSlot.chance) then 
				local item = CreateItem(def.bootsSlot, level, "boots")
				if item then
					hero:SetAttribute("feetItem", item)
					hero:AddChild(item)
				end
			end
		end
		
		if def.equipmentSlot then
			if (not def.equipmentSlot.chance) or (math.random() <= def.equipmentSlot.chance) then 
				local item = CreateItem(def.equipmentSlot, level, "misc")
				if item then
					hero:SetAttribute("jewelleryItem", item)
					hero:AddChild(item)
				end
			end
		end
		
		if def.activeSpells then
			for name, minLevel in pairs(def.activeSpells) do
				if minLevel <= level then
					hero:PushAttribute("activeSpells", name)
				end
			end
		end
		
		if def.knownSpells then
			for i, name in ipairs(def.knownSpells) do
				hero:PushAttribute("knownSpells", name)
			end
		end
		
		if def.inventory then
			for i, itemdef in ipairs(def.inventory) do
				if (not itemdef.chance) or (math.random() <= itemdef.chance) then 
					local item = CreateItem(itemdef, level)
					
					hero:PushAttribute("inventory", item)
					hero:AddChild(item)
				end
			end
		end
		
		if def.traits then
			for _, name in ipairs(def.traits) do
				hero:PushAttribute("traits", name)
			end
		end
		
		if not def.disableLootDrops then
			-- Monster drops are determined by the following design:
			-- 	ITEM					CHANCE			AMOUNT
			--  Wood					(50+Level)%		1 .. ceil(LEVEL/10)
			--  Metal					(50+Level)%		1 .. ceil(LEVEL/10)
			--  Leather					(50+Level)%		1 .. ceil(LEVEL/10)
			--  Group Catalyst			(75+Level)%		1 .. ceil(LEVEL/10)
			--  Random Catalyst			(Level/2)%		1 .. ceil(LEVEL/10)
			--  Gold					10%				50 + 1 .. sqrt(LEVEL) * 100
			
			if math.random(0,100) < (50 + level) then
				--hero:AddCommodity("wood", math.random(1, math.ceil(level / 10)))
				hero:AddCommodity("wood", math.random(1,5))
			end
			
			if math.random(0,100) < (50 + level) then
				--hero:AddCommodity("metal", math.random(1, math.ceil(level / 10)))
				hero:AddCommodity("metal", math.random(1,5))
			end
			
			if math.random(0,100) < (50 + level) then
				--hero:AddCommodity("leather", math.random(1, math.ceil(level / 10)))
				hero:AddCommodity("leather", math.random(1,5))
			end
			
			if def.catalystItem then
				if math.random(0,100) < (75 + level / 2) then
					hero:AddCommodity(def.catalystItem, math.random(1, math.ceil(level / 10)))
				end
			end
			
			local randomCatalyst = table.random(ItemConfiguration.CATALYST_ITEMS)
			
			if math.random(0,100) < (level / 2) or not def.catalystItem or hero:GetCommodityAmount(def.catalystItem) == 0 then
				hero:AddCommodity(randomCatalyst, hero:GetCommodityAmount(randomCatalyst) + math.random(1, math.ceil(level / 20)))
			end
		end
		
		-- [SC] TODO: monsters and hero class definitions are out of sync with each other and the new stats system
		-- change, so we translate the monster definition into minimal fake hero class tables for the stats calculator
		hero:InitializeStats()
		
		local monster_class_base = {
			hit_points_per_level = def.hitPointsPerLevel,
			start_data = {
				health = def.baseHitPoints,
				max_mana = {
					red = def.max_red_mana or 20, 
					green = def.max_green_mana or 20, 
					blue = def.max_blue_mana or 20, 
					yellow = def.max_yellow_mana or 20, 
					black = def.max_black_mana or 20,
				},
			},
		}
		
		-- we incorporate all item effects on top of our stats
		hero:CalculateStats( monster_class_base, nil, level )
		
		-- we add monster specific derived stat bonuses
		hero.stats.derived.defence_rating = hero.stats.derived.defence_rating + ( def.naturalArmour or 0 )
		
		LOG("Monster generated:")
		hero:PrintDebug(true)
		
		return hero
	end)


end

function CreateMonsterHero(monster, referenceHero, nodeID, die)
    assert(monster)
    assert(referenceHero)
    
	return CreateMonsterWithoutReferenceHero(monster, referenceHero:GetAttribute("level"), referenceHero:GetAttribute("difficulty"), nodeID, die)

end

function CreateRandomMonster(referenceHero, quickBattle)
	assert(referenceHero)
	
	if IsPlatformPSP() then -- this is a hack for demo purposes
		return CreateMonsterHero("GelatinousCube", referenceHero)
	else
		local monsterCandidates = GetListOfMonsters(referenceHero, quickBattle)
		
		local listMonsterSize = 0
		for index, mon in pairs(monsterCandidates) do
			listMonsterSize = listMonsterSize + 1
		end
		
		local hero = HeroManager:GetActiveHero()
		
		local lastQuickID = hero:GetAttribute("dummy")
		if not lastQuickID or lastQuickID < 1 or lastQuickID > listMonsterSize then
			lastQuickID = 1
		end
		
		local lastMonster = monsterCandidates[lastQuickID]
		
		return CreateMonsterHero(lastMonster, referenceHero)
	end
end

-- quickBattle indicates that QuickBattleList.lua should be used instead of MonsterList.lua
-- and quest predicates need to be checked to see if bosses should be added.
function GetListOfMonsters(referenceHero, quickBattle) 
	local level = referenceHero:GetAttribute("level")
	
	local monsterCandidates = {}
	
	if quickBattle then
		local monsters = require("Monsters/QuickBattleList")
		for _, monster in ipairs(monsters) do
			local data = MONSTERS[monster.name]
			if data then
				--if data.minLevel <= level and data.maxLevel >= level then
				if data.minLevel <= level then
					if monster.questData then
						if CheckQuestPredicates(referenceHero, monster.questData) then
							table.insert(monsterCandidates, monster.name)
						end
					else
						table.insert(monsterCandidates, monster.name)
					end
				end
			end
		end
	
	else
		local monsters = require("Monsters/MonstersList")
		for _, monster in ipairs(monsters) do
			local data = MONSTERS[monster]
			if data and not data.boss then
				--if data.minLevel <= level and data.maxLevel >= level then
				if data.minLevel <= level then
					table.insert(monsterCandidates, monster)
				end
			end
		end
	end

	return monsterCandidates
end